@@ -140,7 +140,7 @@ void set_initial_parameter_values(Model& M, const variables_map& args)
 	    std::ostringstream o;
 	    o<<"Setting parameter '"<<name<<"': ";
 	    e.prepend(o.str());
-	    throw e;
+	    throw;
 	}
 
 	auto p_index = M.maybe_find_parameter(name);
@@ -401,6 +401,11 @@ int main(int argc,char* argv[])
 	L->beta_reduction = args["beta-reduction"].as<bool>();
 	L->max_iterations = args["simplifier-max-iterations"].as<int>();
 
+	L->fully_lazy = args["fully-lazy"].as<bool>();
+	L->dump_parsed = args.count("dump-parsed");
+	L->dump_renamed = args.count("dump-rn");
+	L->dump_desugared = args.count("dump-ds");
+
 	//---------- Initialize random seed -----------//
 	unsigned long seed = init_rng_and_get_seed(args);
     
@@ -410,7 +415,7 @@ int main(int argc,char* argv[])
 	if (args.count("test-module"))
 	{
 	    string filename = args["test-module"].as<string>();
-	    Module M ( L->read_module_from_file(filename) );
+	    auto M = L->load_module_from_file(filename);
 
 	    Program P(L);
 	    P.add(M);
@@ -423,6 +428,12 @@ int main(int argc,char* argv[])
 	    }
 	    exit(0);
 	}
+	if (args.count("run-module"))
+	{
+	    string filename = args["run-module"].as<string>();
+	    execute_file(L, filename);
+	    exit(0);
+	}
 
 	//---------- Create model object -----------//
 	json info;
@@ -459,7 +470,7 @@ int main(int argc,char* argv[])
 	    model_t print = get_model(R,"a",mstring);
 
 	    expression_ref print_exp = print.expression;
-	    expression_ref a = {var("Prelude.error"),"No alphabet!"};
+	    expression_ref a = error("No alphabet!");
 	    if (args.count("alphabet"))
 	    {
 		auto as = args["alphabet"].as<vector<string>>();
@@ -474,15 +485,18 @@ int main(int argc,char* argv[])
 			a = RNA();
 		    else if (alpha == "AA")
 			a = AminoAcids();
+		    else if (alpha == "Doublets")
+			a = Doublets(RNA());
 		    else if (alpha == "Triplets")
 			a = Triplets(DNA());
 		    else if (alpha == "Codons")
-			a = Codons(DNA(),AminoAcids(),Standard_Genetic_Code());
+			a = Codons(DNA(), AminoAcids(), Standard_Genetic_Code());
 		}
 	    }
-	    print_exp = {var("Distributions.run_random"), a, print_exp};
+	    print_exp = {var("Distributions.run_random"), a, false, print_exp};
 	    print_exp = {var("Prelude.unsafePerformIO"),print_exp};
 	    print_exp = {var("Parameters.evaluate"),-1,print_exp};
+	    print_exp = {var("Data.Tuple.fst"),print_exp };
 	    print_exp = {var("Prelude.show"),print_exp };
 	    print_exp = {var("Prelude.listToString"),print_exp };
 	    int print_exp_index = M->add_compute_expression( print_exp );
@@ -506,7 +520,8 @@ int main(int argc,char* argv[])
 	{
 	    auto P = M.as<Parameters>();
 	    for(int i=0;i<P->n_branch_scales();i++)
-		P->branch_scale(i, 1.0);
+		if (P->branch_scale_modifiable_reg(i))
+		    P->branch_scale(i, 1.0);
 	}
 
 	set_initial_parameter_values(*M,args);
