@@ -1,5 +1,5 @@
 #pragma clang diagnostic ignored "-Wreturn-type-c-linkage"
-#include "computation/computation.H"
+#include "computation/machine/args.H"
 #include "myexception.H"
 #include "computation/machine/graph_register.H"
 #include "rng.H"
