@@ -1,5 +1,5 @@
 #pragma clang diagnostic ignored "-Wreturn-type-c-linkage"
-#include "computation/computation.H"
+#include "computation/machine/args.H"
 #include "myexception.H"
 #include <algorithm>
 #include "computation/expression/expression.H"
