@@ -1,6 +1,6 @@
 #pragma clang diagnostic ignored "-Wreturn-type-c-linkage"
 //#define DEBUG_RATE_MATRIX
-#include "computation/computation.H"
+#include "computation/machine/args.H"
 #include "math/exponential.H"
 #include "math/eigenvalue.H"
 #include "sequence/alphabet.H"
@@ -18,27 +18,10 @@ using std::cerr;
 using std::endl;
 using std::abs;
 
-vector<double> from_evec(const EVector& v1)
-{
-    vector<double> v2(v1.size());
-    for(int i=0;i<v2.size();i++)
-	v2[i] = v1[i].as_double();
-    return v2;
-}
-
-EVector to_evec(const vector<double>& v1)
-{
-    EVector v2(v1.size());
-    for(int i=0;i<v2.size();i++)
-	v2[i] = v1[i];
-    return v2;
-}
-
-
 extern "C" closure builtin_function_lExp(OperationArgs& Args)
 {
     auto L = Args.evaluate(0);
-    auto pi = from_evec( Args.evaluate(1).as_<EVector>() );
+    vector<double> pi = Args.evaluate(1).as_<EVector>();
     double t = Args.evaluate(2).as_double();
 
     auto M = new Box<Matrix>;
@@ -59,7 +42,7 @@ extern "C" closure builtin_function_get_eigensystem(OperationArgs& Args)
     auto arg0 = Args.evaluate(0);
     const Matrix& Q = arg0.as_< Box<Matrix> >();
 
-    auto pi = from_evec(Args.evaluate(1).as_<EVector>() );
+    auto pi = vector<double>(Args.evaluate(1).as_<EVector>() );
 
     const unsigned n = Q.size1();
     assert(Q.size2() == Q.size1());
@@ -119,7 +102,7 @@ extern "C" closure builtin_function_get_equilibrium_rate(OperationArgs& Args)
     auto arg2 = Args.evaluate(2);
     const Matrix& Q = arg2.as_< Box<Matrix> >();
 
-    auto pi = from_evec (Args.evaluate(3).as_<EVector>() );
+    auto pi = vector<double> (Args.evaluate(3).as_<EVector>() );
 
     assert(Q.size2() == Q.size1());
     const unsigned N = smap.size();
@@ -147,49 +130,149 @@ extern "C" closure builtin_function_get_equilibrium_rate(OperationArgs& Args)
     return {scale/a.width()};
 }
 
-
-extern "C" closure builtin_function_singlet_to_triplet_exchange(OperationArgs& Args)
+extern "C" closure builtin_function_rna_16a_exchange(OperationArgs& Args)
 {
     auto arg0 = Args.evaluate(0);
-    const Triplets& T = arg0.as_<Triplets>();
+    const Doublets& D = arg0.as_<Doublets>();
 
-    auto arg1 = Args.evaluate(1);
-    const Matrix& S2 = arg1.as_<Box<Matrix>>();
+    // 2 changes: between matches/wobbles (both transitions)
+    double alpha_D = Args.evaluate(2).as_double();
+    // 2 changes: between matches/wobbles (both transversions)
+    double beta = Args.evaluate(3).as_double();
+    // 2 changes: mismatch and anything: ZERO
 
-    int N = T.size();
+    // 1 change: between matches (must be transition between match and wobble)
+    double alpha_S = Args.evaluate(1).as_double();
+    // 1 change: between match and mismatch
+    double gamma = Args.evaluate(4).as_double();
+    // 1 change: between mismatch and mismatch
+    double epsilon = Args.evaluate(5).as_double();
 
-    object_ptr<Box<Matrix>> R ( new Box<Matrix>(N,N) );
+    const int n = D.size();
 
-    Matrix& S = *R;
+    object_ptr<Box<Matrix>> R( new Box<Matrix>(n,n) );
 
-    for(int i=0;i<T.size();i++)
-	for(int j=0;j<i;j++) 
+    for(int i=0;i<n;i++)
+    {
+	(*R)(i,i) = 0;
+	for(int j=i+1;j<n;j++)
 	{
-	    int nmuts=0;
-	    int pos=-1;
-	    for(int p=0;p<3;p++)
-		if (T.sub_nuc(i,p) != T.sub_nuc(j,p)) {
-		    nmuts++;
-		    pos=p;
+	    if (i==j) continue;
+
+	    bool m1 = D.is_watson_crick(i) or D.is_wobble_pair(i);
+	    bool m2 = D.is_watson_crick(j) or D.is_wobble_pair(j);
+
+	    double x = 0;
+	    if (D.n_changes(i,j) == 2)
+	    {
+		if (m1 and m2)
+		{
+		    int i0 = D.sub_nuc(i,0);
+		    int i1 = D.sub_nuc(i,1);
+
+		    int j0 = D.sub_nuc(j,0);
+		    int j1 = D.sub_nuc(j,1);
+
+		    auto& N = D.getNucleotides();
+		    if (N.transition(i0,j0))
+		    {
+			// double transition -> alpha_D
+			x = alpha_D;
+			assert(N.transition(i1,j1));
+		    }
+		    else
+		    {
+			// double transversion -> beta
+			x = beta;
+			assert(N.transversion(i1,j1));
+		    }
+		}
+		else {
+		    // double change between mismatch and something
+		    x = 0;
 		}
-	    assert(nmuts>0);
-	    assert(pos >= 0 and pos < 3);
-	
-	    S(j,i) = S(i,j) = 0;
-
-	    if (nmuts == 1) {
-
-		int l1 = T.sub_nuc(i,pos);
-		int l2 = T.sub_nuc(j,pos);
-		assert(l1 != l2);
-
-		S(j,i) = S(i,j) = S2(l1,l2);
 	    }
+	    else if (D.n_changes(i,j) == 1)
+	    {
+		int n = (m1?1:0) + (m2?1:0);
+		//    match <->    match = alpha_S
+		if (n == 2)
+		    x = alpha_S;
+		// mismatch <->    match = gamma
+		else if (n==1)
+		    x = gamma;
+		// mismatch <-> mismatch = epsilon
+		else
+		    x = epsilon;
+	    }
+
+	    (*R)(i,j) = (*R)(j,i) = x;
 	}
+    }
 
     return R;
 }
 
+extern "C" closure builtin_function_singlet_to_doublet_rates(OperationArgs& Args)
+{
+    auto arg0 = Args.evaluate(0);
+    const Doublets& D = arg0.as_<Doublets>();
+
+    auto arg1 = Args.evaluate(1);
+    const Matrix& R1 = arg1.as_<Box<Matrix>>();
+
+    auto arg2 = Args.evaluate(2);
+    const Matrix& R2 = arg2.as_<Box<Matrix>>();
+
+    // The way alphabet is currently implemented, doublets must be doublets of nucleotides.
+    assert(R1.size1() == 4);
+    assert(R1.size2() == 4);
+    assert(R2.size1() == 4);
+    assert(R2.size2() == 4);
+
+    const int n = D.size();
+
+    object_ptr<Box<Matrix>> R( new Box<Matrix>(n,n) );
+
+    for(int i=0;i<n;i++)
+    {
+	double sum = 0;
+	for(int j=0;j<n;j++)
+	{
+	    if (i==j) continue;
+	    int nmuts=0;
+	    int from=-1;
+	    int to=-1;
+	    int pos=-1;
+	    for(int p=0;p<2;p++)
+		if (D.sub_nuc(i,p) != D.sub_nuc(j,p))
+		{
+		    nmuts++;
+		    pos = p;
+		    from = D.sub_nuc(i,p);
+		    to = D.sub_nuc(j,p);
+		}
+
+	    double r = 0;
+	    if (nmuts == 1)
+	    {
+		if (pos == 0)
+		    r = R1(from,to);
+		else if (pos == 1)
+		    r = R2(from,to);
+		else
+		    std::abort();
+	    }
+	    (*R)(i,j) = r;
+	    sum += r;
+	}
+	(*R)(i,i) = -sum;
+    }
+
+    return R;
+}
+
+
 extern "C" closure builtin_function_singlet_to_triplet_rates(OperationArgs& Args)
 {
     auto arg0 = Args.evaluate(0);
@@ -698,6 +781,47 @@ extern "C" closure builtin_function_f3x4_frequencies(OperationArgs& Args)
     return pi;
 }
 
+extern "C" closure builtin_function_f2x4_frequencies(OperationArgs& Args)
+{
+    auto arg0 = Args.evaluate(0);
+    const Doublets& D = arg0.as_<Doublets>();
+    // The way alphabet is currently implemented, triplets must be triplets of nucleotides.
+
+    auto arg1 = Args.evaluate(1);
+    auto pi1 = arg1.as_<EVector>();
+
+    int nuc_size = D.getNucleotides().size();
+
+    if (pi1.size() != nuc_size)
+	throw myexception()<<"f2x4_frequencies:site 1:expected "<<nuc_size<<" frequencies, but got "<<pi1.size()<<"!";
+
+    auto arg2 = Args.evaluate(2);
+    auto pi2 = arg2.as_<EVector>();
+
+    if (pi2.size() != nuc_size)
+	throw myexception()<<"f2x4_frequencies:site 2:expected "<<nuc_size<<" frequencies, but got "<<pi2.size()<<"!";
+
+    EVector pi;
+    pi.resize(D.size());
+    double sum = 0;
+    for(int i=0;i<D.size();i++)
+    {
+	double x = pi1[D.sub_nuc(i,0)].as_double() * pi2[D.sub_nuc(i,1)].as_double();
+	pi[i] = x;
+	sum += x;
+    }
+
+    // Some triplets may be missing from the doublet alphabet (e.g. mismatches). So renormalize
+
+    double scale = 1.0/sum;
+    for(auto& d : pi)
+	d = d.as_double() * scale;
+
+//    assert(std::abs(sum(pi) - 1.0) < 1.0e-9);
+
+    return pi;
+}
+
 extern "C" closure builtin_function_gtr_sym(OperationArgs& Args)
 {
     auto arg0 = Args.evaluate(0);
@@ -850,7 +974,7 @@ extern "C" closure builtin_function_plus_gwF(OperationArgs& Args)
 
     double f = Args.evaluate(1).as_double();
 
-    auto pi = from_evec( Args.evaluate(2).as_<EVector>() );
+    auto pi = vector<double>( Args.evaluate(2).as_<EVector>() );
 
     const int n = a.size();
 
@@ -876,74 +1000,51 @@ extern "C" closure builtin_function_plus_gwF(OperationArgs& Args)
     return R;
 }
 
-// codon_a codon_w omega nuc_q
-extern "C" closure builtin_function_fMutSel_q(OperationArgs& Args)
+double bound(double low, double high, double x)
+{
+    if (x < low) return low;
+    if (x > high) return high;
+    return x;
+}
+
+// Q0 w
+// Here S[I,J] = F[J] - F[I] = 2Nf[j] - 2NF[i] = 2N*s[i,j]
+extern "C" closure builtin_function_mut_sel_q(OperationArgs& Args)
 {
     auto arg0 = Args.evaluate(0);
-    const Codons& C = arg0.as_<Codons>();
-    int N = C.size();
-  
-    auto arg1 = Args.evaluate(1);
-    const Vector<double>& codon_w = arg1.as_< Vector<double> >();
-    assert(codon_w.size() == N);
+    const Matrix& Q0 = arg0.as_< Box<Matrix> >();
+    assert(Q0.size1() == Q0.size2());
+    int n = Q0.size1();
 
-    double omega = Args.evaluate(2).as_double();
+    auto F   = vector<double>( Args.evaluate(1).as_< EVector >() );
+    for(auto& f: F)
+	f = bound(-20,20,f);
 
-    auto arg3 = Args.evaluate(3);
-    const Matrix& nuc_Q = arg3.as_< Box<Matrix> >();
-    assert(nuc_Q.size1() == nuc_Q.size2());
-    assert(nuc_Q.size1() == C.getNucleotides().size());
+    assert(F.size() == n);
 
-    vector<double> log_codon_w(N);
-    for(int i=0;i<N;i++)
-	log_codon_w[i] = log(codon_w[i]);
-  
-    auto Q_ = new Box<Matrix>(N,N);
+    auto Q_ = new Box<Matrix>(n,n);
     Matrix& Q = *Q_;
 
-    for(int i=0;i<N;i++)
+    for(int i=0;i<n;i++)
     {
 	double sum = 0;
-	for(int j=0;j<N;j++)
+	for(int j=0;j<n;j++)
 	{
 	    if (i==j) continue;
 
-	    int nmuts=0;
-	    int pos=-1;
-	    for(int p=0;p<3;p++)
-		if (C.sub_nuc(i,p) != C.sub_nuc(j,p)) {
-		    nmuts++;
-		    pos=p;
-		}
-	    assert(nmuts>0);
-	    assert(pos >= 0 and pos < 3);
+	    double rate = Q0(i,j);
 
-	    double rate=0.0;
-
-	    if (nmuts == 1) 
-	    {
-		int l1 = C.sub_nuc(i,pos);
-		int l2 = C.sub_nuc(j,pos);
-		assert(l1 != l2);
-
-		rate = nuc_Q(l1,l2);
-
-
-		// x = wj/wi    log(x)/(1-1/x)
-		// y = wi/wj   -log(y)/(1-y)
-		// 1+z = y     -log(1+z)/-z = log1p(z)/z   z = y-1 = (wi/wj)-1
-		double z = (codon_w[i] - codon_w[j])/codon_w[j];
-		if (std::abs(z) < 0.0001)
-		    rate *= ( 1.0 - z/2.0 + (z*z)/3.0 - (z*z*z)/4.0 );
-		else
-		    rate *= log1p(z)/z;
-
-		if (C.translate(i) != C.translate(j))
-		    rate *= omega;	
-	    }
+	    // x = wj/wi    log(x)/(1-1/x)
+	    // y = wi/wj   -log(y)/(1-y)
+	    // 1+z = y     -log(1+z)/-z = log1p(z)/z   z = y-1 = (wi/wj)-1
+	    double S = F[j] - F[i];
+	    if (std::abs(S) < 0.0001)
+		rate *= ( 1.0 + S/2 + (S*S)/12 - (S*S*S*S)/720 );
+	    else
+		rate *= -S/expm1(-S);
 
 	    Q(i,j) = rate;
-      
+
 	    sum += Q(i,j);
 	}
 	Q(i,i) = -sum;
@@ -952,28 +1053,27 @@ extern "C" closure builtin_function_fMutSel_q(OperationArgs& Args)
     return Q_;
 }
 
-// codon_a nuc_pi codon_w
-extern "C" closure builtin_function_fMutSel_pi(OperationArgs& Args)
+// pi0 w
+extern "C" closure builtin_function_mut_sel_pi(OperationArgs& Args)
 {
-    auto arg0 = Args.evaluate(0);
-    const Codons& C = arg0.as_<Codons>();
-    int N = C.size();
-  
-    auto arg1 = Args.evaluate(1);
-    const vector<double>& codon_w = arg1.as_< Vector<double> >();
-    assert(codon_w.size() == N);
+    auto pi0 = vector<double>( Args.evaluate(0).as_< EVector >() );
 
-    auto nuc_pi = from_evec( Args.evaluate(2).as_< EVector >() );
-    assert(nuc_pi.size() == C.getNucleotides().size());
+    auto F   = vector<double>( Args.evaluate(1).as_< EVector >() );
+    for(auto& f: F)
+	f = bound(-20,20,f);
+
+    assert(pi0.size() == F.size());
 
     // compute frequencies
-    Vector<double> pi(C.size());
-    
-    for(int i=0;i<N;i++)
-	pi[i] = nuc_pi[C.sub_nuc(i,0)] * nuc_pi[C.sub_nuc(i,1)] * nuc_pi[C.sub_nuc(i,2)] * codon_w[i];
+    vector<double> pi = pi0;
+
+    double Fmax = max(F);
+
+    for(int i=0; i<pi.size(); i++)
+	pi[i] *= exp(F[i]-Fmax);
 
     normalize(pi);
-    return to_evec(pi);
+    return EVector(pi);
 }
 
 /*
