@@ -89,11 +89,6 @@ long max_version = 0;
  *    loop until no more regs (and thus computations) are being freed.
  */
 
-bool null(const CacheList<int>::iterator& i)
-{
-    return (i == CacheList<int>::iterator());
-}
-
 void Step::clear()
 {
     source_reg = -1;
@@ -187,14 +182,17 @@ reg& reg::operator=(reg&& R) noexcept
 
     created_by = std::move(R.created_by);
 
+    flags = R.flags;
+
     return *this;
 }
 
 reg::reg(reg&& R) noexcept
 :C( std::move(R.C) ),
-			 type ( R.type ),
-			 n_heads( R.n_heads ),
-			 created_by( std::move(R.created_by) )
+ type ( R.type ),
+ n_heads( R.n_heads ),
+ created_by( std::move(R.created_by) ),
+ flags ( R.flags )
 { }
 
 void reg::clear()
@@ -202,6 +200,7 @@ void reg::clear()
     assert(n_heads == 0);
     C.clear();
     type = type_t::unknown;
+    flags.reset();
 }
 
 void reg::check_cleared()
@@ -210,7 +209,8 @@ void reg::check_cleared()
     assert(type == type_t::unknown);
     assert(n_heads == 0);
     assert(created_by.first == 0);
-    assert(null(created_by.second));
+    assert(created_by.second == 0);
+    assert(flags.none());
 }
 
 void mapping::add_value(int r, int v) 
@@ -247,134 +247,380 @@ bool mapping::empty() const
     return delta_.empty();
 }
 
-void reg_heap::register_probability(int r)
+boost::optional<int> reg_heap::creator_of_reg(int r) const
+{
+    int s = regs[r].created_by.first;
+    assert(s >= 0);
+    if (s == 0)
+	return boost::none;
+    else
+	return s;
+}
+
+bool reg_heap::reg_is_contingent(int r) const
+{
+    return (bool)creator_of_reg(r);
+}
+
+bool reg_heap::step_exists_in_root(int s) const
+{
+    assert(s > 0);
+    int r = steps[s].source_reg;
+    assert(r > 0 and r < size());
+    return prog_steps[r] == s;
+}
+
+bool reg_heap::reg_exists(int r) const
+{
+    auto s = creator_of_reg(r);
+    if (not s)
+	return true;
+    else
+	return step_exists_in_root(*s);
+}
+
+size_t reg_heap::size() const
+{
+    assert(regs.size() == prog_steps.size());
+    assert(regs.size() == prog_results.size());
+    assert(regs.size() == prog_temp.size());
+    return regs.size();
+}
+
+void reg_heap::register_prior(int r)
 {
     mark_completely_dirty(root_token);
-    r = incremental_evaluate(r).first;
+    r = incremental_evaluate_unchangeable(r);
 
     if (reg_is_constant(r))
     {
-	log_double_t pr = access(r).C.exp.as_log_double();
-	constant_pr *= pr;
+	log_double_t pr = regs.access(r).C.exp.as_log_double();
+	constant_prior *= pr;
     }
     else
     {
+	regs.access(r).flags.set(0);
+
 	assert(reg_is_changeable(r));
 
-	int rc = result_index_for_reg(r);
-	assert(rc > 0);
+	prior_heads.push_back(r);
 
-	probability_heads.push_back(r);
-
-	prs_list.push_back(r);
+	priors_list.push_back(r);
     }
 }
 
-int reg_heap::register_probability(closure&& C)
+int reg_heap::register_prior(closure&& C)
 {
     assert(not C.exp.head().is_a<expression>());
 
     int r = allocate();
-    total_reg_allocations++;
     set_C(r, std::move(C));
-    register_probability(r);
+    register_prior(r);
     return r;
 }
 
-bool reg_heap::inc_probability_for_reg(int r)
+bool reg_heap::inc_prior_for_reg(int r)
 {
     assert(reg_is_changeable(r));
     int rc = result_index_for_reg(r);
 
+    if (not reg_exists(r)) return false; // reg does not exist in this context!
+
     if (rc > 0 and results[rc].flags.test(0)) return true; // already included
 
     incremental_evaluate(r);
     rc = result_index_for_reg(r);
 
-    return inc_probability(rc);
+    return inc_prior(rc);
 }
 
-void reg_heap::dec_probability_for_reg(int r)
+void reg_heap::dec_prior_for_reg(int r)
 {
     int rc = result_index_for_reg(r);
 
     if (rc > 0 and results[rc].flags.test(0))
-	dec_probability(rc);
+	dec_prior(rc);
 }
 
-void reg_heap::dec_probability(int rc)
+void reg_heap::dec_prior(int rc)
 {
     assert(rc > 0);
     int r2 = results[rc].value;
     assert(r2 > 0);
-    log_double_t pr = access(r2).C.exp.as_log_double();
+    log_double_t pr = regs.access(r2).C.exp.as_log_double();
+
+    // This value has already by included, so take it out unconditionally.
+    prior.data.value -= pr.log();
 
-    variable_pr /= pr;
     results[rc].flags.reset(0);
 
     int r = results[rc].source_reg;
     assert(reg_is_changeable(r));
-    prs_list.push_back(r);
+    priors_list.push_back(r);
 }
 
-double id(double x) {return x;}
-
-log_double_t reg_heap::probability_for_context_diff(int c)
+log_double_t reg_heap::prior_for_context_diff(int c)
 {
     reroot_at_context(c);
-    unhandled_pr.log() = 0.0;
+    prior.reset_unhandled();
 
     // re-multiply all probabilities
-    if (total_error > 1.0e-9)
+    if (prior.data.total_error > 1.0e-9)
     {
-	for(int r: probability_heads)
+	for(int r: prior_heads)
 	{
 	    int rc = result_index_for_reg(r);
 	    if (rc > 0 and results[rc].flags.test(0))
-		dec_probability(rc);
+		dec_prior(rc);
 	}
-	// std::cerr<<"unwinding all prs: total_error = "<<total_error<<" variable_pr = "<<variable_pr<<"  error_pr = "<<error_pr<<"   variable_pr/error_pr = "<<variable_pr/error_pr<<std::endl;
-	assert((variable_pr/error_pr).log() < 1.0e-6);
-	assert(prs_list.size() == probability_heads.size());
-	total_error = 0;
-	variable_pr.log() = 0;
-	error_pr.log() = 0;
+	// std::cerr<<"unwinding all prs: total_error = "<<prior.data.total_error<<" variable_pr = "<<prior.data.value<<"  error_pr = "<<prior.data.delta<<"   variable_pr/error_pr = "<<prior.data.value - prior.data.delta<<std::endl;
+	assert(std::abs(prior.data.value - prior.data.delta) < 1.0e-6);
+	assert(priors_list.size() == prior_heads.size());
+	prior.reset();
     }
 
-    if (not prs_list.empty())
+    if (not priors_list.empty())
     {
 	mark_completely_dirty(root_token);
 
 	int j=0;
-	for(int i=0;i<prs_list.size();i++)
+	for(int i=0;i<priors_list.size();i++)
 	{
-	    int r = prs_list[i];
-	    if (not inc_probability_for_reg(r))
-		prs_list[j++] = r;
+	    int r = priors_list[i];
+	    if (not inc_prior_for_reg(r))
+		priors_list[j++] = r;
 	}
-	prs_list.resize(j);
+	priors_list.resize(j);
     }
 
-    return variable_pr * constant_pr * unhandled_pr / error_pr;
+    return constant_prior * log_double_t(prior);
 }
 
-log_double_t reg_heap::probability_for_context(int c)
+log_double_t reg_heap::prior_for_context(int c)
 {
     total_context_pr++;
 
-    log_double_t Pr = probability_for_context_diff(c);
-    // std::cerr<<"A:   Pr1 = "<<Pr<<"   error = "<<total_error<<"  constant_pr = "<<constant_pr<<"  variable_pr = "<<variable_pr<<"  unhandled = "<<unhandled_pr<<std::endl;
+    log_double_t Pr = prior_for_context_diff(c);
+    // std::cerr<<"A:   Pr1 = "<<Pr<<"   error = "<<prior.data.total_error<<"  constant_pr = "<<constant_prior<<"  variable_pr = "<<prior.data.value<<"  unhandled = "<<prior.data.unhandled<<std::endl;
 
 #ifndef NDEBUG  
-    // log_double_t Pr2 = probability_for_context_full(c);
+    // log_double_t Pr2 = prior_for_context_full(c);
     // double diff = Pr.log() - Pr2.log();
-    // std::cerr<<"B:diff = "<<diff<<"    Pr1 = "<<Pr<<"  Pr2 = "<<Pr2<<"   error = "<<total_error<<"  constant_pr = "<<constant_pr<<"  variable_pr = "<<variable_pr<<"  unhandled = "<<unhandled_pr<<std::endl;
+    // std::cerr<<"B:diff = "<<diff<<"    Pr1 = "<<Pr<<"  Pr2 = "<<Pr2<<"   error = "<<prior.data.total_error<<"  constant_pr = "<<constant_prior<<"  variable_pr = "<<prior.data.value<<"  unhandled = "<<prior.data.unhandled<<std::endl;
     //  assert(fabs(diff) < 1.0e-6);
 #endif
 
     return Pr;
 }
 
+void reg_heap::register_likelihood(int r)
+{
+    mark_completely_dirty(root_token);
+    r = incremental_evaluate_unchangeable(r);
+
+    if (reg_is_constant(r))
+    {
+	log_double_t pr = regs.access(r).C.exp.as_log_double();
+	constant_likelihood *= pr;
+    }
+    else
+    {
+	regs.access(r).flags.set(1);
+
+	assert(reg_is_changeable(r));
+
+	likelihood_heads.push_back(r);
+
+	likelihoods_list.push_back(r);
+    }
+}
+
+int reg_heap::register_likelihood(closure&& C)
+{
+    assert(not C.exp.head().is_a<expression>());
+
+    int r = allocate();
+    set_C(r, std::move(C));
+    register_likelihood(r);
+    return r;
+}
+
+bool reg_heap::inc_likelihood_for_reg(int r)
+{
+    assert(reg_is_changeable(r));
+    int rc = result_index_for_reg(r);
+
+    assert(reg_exists(r)); // reg had better exist in this context!
+
+    if (rc > 0 and results[rc].flags.test(1)) return true; // already included
+
+    incremental_evaluate(r);
+    rc = result_index_for_reg(r);
+
+    return inc_likelihood(rc);
+}
+
+void reg_heap::dec_likelihood_for_reg(int r)
+{
+    int rc = result_index_for_reg(r);
+
+    if (rc > 0 and results[rc].flags.test(1))
+	dec_likelihood(rc);
+}
+
+void reg_heap::dec_likelihood(int rc)
+{
+    assert(rc > 0);
+    int r2 = results[rc].value;
+    assert(r2 > 0);
+    log_double_t pr = regs.access(r2).C.exp.as_log_double();
+
+    // This value has already by included, so take it out unconditionally.
+    likelihood.data.value -= pr.log();
+
+    results[rc].flags.reset(1);
+
+    int r = results[rc].source_reg;
+    assert(reg_is_changeable(r));
+    likelihoods_list.push_back(r);
+}
+
+log_double_t reg_heap::likelihood_for_context_diff(int c)
+{
+    reroot_at_context(c);
+    likelihood.reset_unhandled();
+
+    // re-multiply all probabilities
+    if (likelihood.data.total_error > 1.0e-9)
+    {
+	for(int r: likelihood_heads)
+	{
+	    int rc = result_index_for_reg(r);
+	    if (rc > 0 and results[rc].flags.test(1))
+		dec_likelihood(rc);
+	}
+	// std::cerr<<"unwinding all prs: total_error = "<<likelihood.data.total_error<<" variable_pr = "<<likelihood.data.value<<"  error_pr = "<<likelihood.data.delta<<"   variable_pr/error_pr = "<<likelihood.data.value - likelihood.data.delta<<std::endl;
+	assert(std::abs(likelihood.data.value - likelihood.data.delta) < 1.0e-6);
+	assert(likelihoods_list.size() == likelihood_heads.size());
+	likelihood.reset();
+    }
+
+    if (not likelihoods_list.empty())
+    {
+	mark_completely_dirty(root_token);
+
+	int j=0;
+	for(int i=0;i<likelihoods_list.size();i++)
+	{
+	    int r = likelihoods_list[i];
+	    if (not inc_likelihood_for_reg(r))
+		likelihoods_list[j++] = r;
+	}
+	likelihoods_list.resize(j);
+    }
+
+    return constant_likelihood * log_double_t(likelihood);
+}
+
+log_double_t reg_heap::likelihood_for_context(int c)
+{
+    total_context_pr++;
+
+    log_double_t Pr = likelihood_for_context_diff(c);
+    // std::cerr<<"A:   Pr1 = "<<Pr<<"   error = "<<likelihood.data.total_error<<"  constant_pr = "<<constant_likelihood<<"  variable_pr = "<<likelihood.data.value<<"  unhandled = "<<likelihood.data.unhandled<<std::endl;
+
+#ifndef NDEBUG  
+    // log_double_t Pr2 = likelihood_for_context_full(c);
+    // double diff = Pr.log() - Pr2.log();
+    // std::cerr<<"B:diff = "<<diff<<"    Pr1 = "<<Pr<<"  Pr2 = "<<Pr2<<"   error = "<<likelihood.data.total_error<<"  constant_pr = "<<constant_likelihood<<"  variable_pr = "<<likelihood.data.value<<"  unhandled = "<<likelihood.data.unhandled<<std::endl;
+    //  assert(fabs(diff) < 1.0e-6);
+#endif
+
+    return Pr;
+}
+
+log_double_t reg_heap::probability_for_context(int c)
+{
+    return prior_for_context(c) * likelihood_for_context(c);
+}
+
+prob_ratios_t reg_heap::probability_ratios(int c1, int c2)
+{
+#if DEBUG_MACHINE >= 2
+    for(auto x : prog_temp)
+	assert(x.none());
+#endif
+
+    // 1. reroot to c1 and force the program
+    /* auto pr1 = */ probability_for_context(c1);
+
+    // 2. install another reroot handler
+    vector<pair<int,int>> original_pdf_results;
+
+    std::function<void(int)> handler = [&original_pdf_results,this](int old_root)
+    {
+	for(auto& p: tokens[old_root].delta_result())
+	{
+	    int r =  p.first;
+
+	    // We're only interested in cases where both contexts have a result that is > 0.
+	    // But (i) we need to seen the "seen" flag in any case
+	    //    (ii) we need to remember that we have set it so that we can unset it.
+	    if (regs.is_used(r) and regs.access(r).flags.any() and not prog_temp[r].test(2))
+	    {
+		prog_temp[r].set(2);
+		original_pdf_results.push_back(p);
+	    }
+	}
+    };
+
+    reroot_handlers.push_back(handler);
+
+    // 3. reroot to c2 and force the program
+    /* auto pr2 = */ probability_for_context(c2);
+
+    // 4. compute the ratio only for (i) changed pdfs that (ii) exist in both c1 and c2
+    prob_ratios_t R{1.0, 1.0, false};
+
+    for(auto& p: original_pdf_results)
+    {
+	int pdf_reg = p.first;
+	int rc1 = p.second;
+
+	assert(prog_temp[pdf_reg].test(2));
+
+	prog_temp[pdf_reg].reset(2);
+
+	// Only compute a ratio if the pdf is present and computed in BOTH contexts.
+	if (rc1 > 0 and has_result(pdf_reg))
+	{
+	    int result_reg1 = results[rc1].value;
+	    int result_reg2 = result_value_for_reg(pdf_reg);
+	    log_double_t r = (*this)[result_reg2].exp.as_log_double() / (*this)[result_reg1].exp.as_log_double();
+
+	    assert(regs.access(pdf_reg).flags.test(0) or regs.access(pdf_reg).flags.test(1));
+	    if (regs.access(pdf_reg).flags.test(0))
+		R.prior_ratio *= r;
+	    else 
+		R.likelihood_ratio *= r;
+	}
+    }
+
+#if DEBUG_MACHINE >= 2
+    for(auto x : prog_temp)
+	assert(x.none());
+#endif
+
+    // 5. remove the reroot handler
+    reroot_handlers.pop_back();
+
+//  If pr1 and pr2 are off far enough, this test will fail...
+//    if (pr1 > 0.0 and pr2 > 0.0)
+//	assert( std::abs( (pr2/pr1).log() - R.prior_ratio.log() - R.likelihood_ratio.log()) < 1.0e-4 );
+
+    return R;
+}
+
 const vector<int>& reg_heap::random_modifiables() const
 {
     return random_modifiables_;
@@ -430,7 +676,7 @@ bool reg_heap::find_modifiable_reg(int& R)
     // Note: here we always update R
     R = incremental_evaluate_unchangeable(R);
 
-    return is_modifiable(access(R).C.exp);
+    return is_modifiable(regs.access(R).C.exp);
 }
 
 const expression_ref reg_heap::get_parameter_range(int c, int p)
@@ -440,30 +686,32 @@ const expression_ref reg_heap::get_parameter_range(int c, int p)
 
 const expression_ref reg_heap::get_range_for_reg(int c, int r)
 {
-    if (access(r).C.Env.size() < 3)
+    if (regs.access(r).C.Env.size() < 3)
 	return {};
 
-    int r2 = access(r).C.lookup_in_env(2);
+    int r2 = regs.access(r).C.lookup_in_env(2);
     return get_reg_value_in_context(r2,c);
 }
 
 double reg_heap::get_rate_for_reg(int r)
 {
-    if (access(r).C.Env.size() < 3)
+    if (regs.access(r).C.Env.size() < 3)
 	return {};
 
-    int r3 = access(r).C.lookup_in_env(0);
+    int r3 = regs.access(r).C.lookup_in_env(0);
     r3 = incremental_evaluate_unchangeable(r3);
-    return access(r3).C.exp.as_double();
+    return regs.access(r3).C.exp.as_double();
 }
 
 int reg_heap::step_index_for_reg(int r) const 
 {
+    assert(prog_steps[r] != 0);
     return prog_steps[r];
 }
 
 int reg_heap::result_index_for_reg(int r) const 
 {
+    assert(prog_results[r] != 0);
     return prog_results[r];
 }
 
@@ -495,12 +743,12 @@ const closure& reg_heap::access_value_for_reg(int R1) const
 {
     int R2 = value_for_reg(R1);
     assert(R2);
-    return access(R2).C;
+    return regs.access(R2).C;
 }
 
 bool reg_heap::reg_has_value(int r) const
 {
-    if (access(r).type == reg::type_t::constant)
+    if (regs.access(r).type == reg::type_t::constant)
 	return true;
     else
 	return reg_has_result_value(r);
@@ -533,12 +781,12 @@ bool reg_heap::has_result(int r) const
 
 int reg_heap::value_for_reg(int r) const 
 {
-    assert(not access(r).C.exp.is_index_var());
-    if (access(r).type == reg::type_t::changeable)
+    assert(not regs.access(r).C.exp.is_index_var());
+    if (regs.access(r).type == reg::type_t::changeable)
 	return result_value_for_reg(r);
     else
     {
-	assert(access(r).type == reg::type_t::constant);
+	assert(regs.access(r).type == reg::type_t::constant);
 	return r;
     }
 }
@@ -559,44 +807,46 @@ void reg_heap::set_result_value_for_reg(int r1)
     assert(value);
 
     int rc1 = result_index_for_reg(r1);
-    if (rc1 <= 0)
+    if (rc1 < 0)
 	rc1 = add_shared_result(r1, step_index_for_reg(r1));
     assert(rc1 > 0);
     auto& RC1 = results[rc1];
     RC1.value = value;
 
     // If R2 is WHNF then we are done
-    if (access(call).type == reg::type_t::constant) return;
+    if (regs.access(call).type == reg::type_t::constant) return;
 
     // If R2 doesn't have a result, add one to hold the called-by edge.
     assert(has_result(call));
 
     // Add a called-by edge to R2.
     int rc2 = result_index_for_reg(call);
-    auto back_edge = results[rc2].called_by.push_back(rc1);
-    RC1.call_edge.first = rc2;
-    RC1.call_edge.second = back_edge;
+    int back_index = results[rc2].called_by.size();
+    results[rc2].called_by.push_back(rc1);
+    RC1.call_edge = {rc2, back_index};
 }
 
 void reg_heap::set_used_input(int s1, int R2)
 {
     assert(reg_is_changeable(R2));
 
-    assert(is_used(R2));
+    assert(regs.is_used(R2));
 
-    assert(access(R2).C);
+    assert(regs.access(R2).C);
 
     assert(has_result(R2));
     assert(result_value_for_reg(R2));
 
     // An index_var's value only changes if the thing the index-var points to also changes.
     // So, we may as well forbid using an index_var as an input.
-    assert(access(R2).C.exp.head().type() != index_var_type);
+    assert(regs.access(R2).C.exp.head().type() != index_var_type);
 
     int rc2 = result_index_for_reg(R2);
 
-    auto back_edge = results[rc2].used_by.push_back(s1);
-    steps[s1].used_inputs.push_back({rc2,back_edge});
+    int back_index = results[rc2].used_by.size();
+    int forw_index = steps[s1].used_inputs.size();
+    results[rc2].used_by.push_back({s1,forw_index});
+    steps[s1].used_inputs.push_back({rc2,back_index});
 
     assert(result_is_used_by(s1,rc2));
 }
@@ -607,10 +857,10 @@ void reg_heap::set_call(int R1, int R2)
     // R2 might be of UNKNOWN changeableness
 
     // Check that R1 is legal
-    assert(is_used(R1));
+    assert(regs.is_used(R1));
 
     // Check that R2 is legal
-    assert(is_used(R2));
+    assert(regs.is_used(R2));
 
     // Only modify the call for the current context;
     assert(has_step(R1));
@@ -643,33 +893,56 @@ void reg_heap::set_C(int R, closure&& C)
     assert(not C.exp.head().is_a<expression>());
     clear_C(R);
 
-    access(R).C = std::move(C);
+    regs.access(R).C = std::move(C);
 #ifndef NDEBUG
-    for(int r: access(R).C.Env)
-	assert(is_valid_address(r));
+    for(int r: regs.access(R).C.Env)
+	assert(regs.is_valid_address(r));
 #endif
 }
 
 void reg_heap::clear_C(int R)
 {
-    truncate(access_unused(R).C);
+    truncate(regs.access_unused(R).C);
 }
 
 void reg_heap::mark_reg_created_by_step(int r, int s)
 {
     assert(r > 0);
     assert(s > 0);
-    steps[s].created_regs.push_front(r);
-    assert(access(r).created_by.first == 0);
-    assert(null(access(r).created_by.second));
-    access(r).created_by = {s,steps[s].created_regs.begin()};
+
+    int index = steps[s].created_regs.size();
+    steps[s].created_regs.push_back(r);
+    assert(regs.access(r).created_by.first == 0);
+    assert(regs.access(r).created_by.second == 0);
+    regs.access(r).created_by = {s,index};
 }
 
-int reg_heap::create_reg_from_step(int s)
+int reg_heap::allocate()
 {
     total_reg_allocations++;
+    return regs.allocate();
+}
+
+int reg_heap::allocate_reg_from_step(int s)
+{
     int r = allocate();
     mark_reg_created_by_step(r,s);
+    assert(not has_step(r));
+    return r;
+}
+
+int reg_heap::allocate_reg_from_step(int s, closure&& C)
+{
+    int r = allocate_reg_from_step(s);
+    set_C(r, std::move(C));
+    return r;
+}
+
+int reg_heap::allocate_reg_from_step_in_token(int s, int t)
+{
+    int r = allocate_reg_from_step(s);
+    tokens[t].vm_result.add_value(r, non_computed_index);
+    tokens[t].vm_step.add_value(r, non_computed_index);
     return r;
 }
 
@@ -698,7 +971,7 @@ void reg_heap::set_reg_value(int R, closure&& value, int t)
     // assert(not is_root_token and tokens[t].version < tokens[parent_token(t)].version) 
 
     // Check that this reg is indeed settable
-    assert(is_modifiable(access(R).C.exp));
+    assert(is_modifiable(regs.access(R).C.exp));
 
     assert(not is_root_token(t));
 
@@ -709,7 +982,7 @@ void reg_heap::set_reg_value(int R, closure&& value, int t)
     tokens[t].vm_step.add_value(R,s);
 
     assert(tokens[t].vm_result.empty());
-    tokens[t].vm_result.add_value(R,-1);
+    tokens[t].vm_result.add_value(R, non_computed_index);
 
     assert(not children_of_token(t).size());
 
@@ -724,7 +997,7 @@ void reg_heap::set_reg_value(int R, closure&& value, int t)
 
 	int Q = value.lookup_in_env( index );
 
-	assert(is_used(Q));
+	assert(regs.is_used(Q));
 
 	// Set the call
 	steps[s].call = Q;
@@ -732,7 +1005,7 @@ void reg_heap::set_reg_value(int R, closure&& value, int t)
     // Otherwise, regardless of whether the expression is WHNF or not, create a new reg for the value and call it.
     else
     {
-	int R2 = create_reg_from_step(s);
+	int R2 = allocate_reg_from_step_in_token(s,t);
 
 	// clear 'reg created' edge from s to old call.
 	steps[s].call = R2;
@@ -821,11 +1094,11 @@ std::vector<int> reg_heap::used_regs_for_reg(int r) const
 void reg_heap::reclaim_used(int r)
 {
     // Mark this reg as not used (but not free) so that we can stop worrying about upstream objects.
-    assert(not access(r).created_by.first);
-    assert(null(access(r).created_by.second));
+    assert(not regs.access(r).created_by.first);
+    assert(not regs.access(r).created_by.second);
     assert(not has_step(r));
   
-    pool<reg>::reclaim_used(r);
+    regs.reclaim_used(r);
 }
 
 template <typename T>
@@ -839,7 +1112,8 @@ void reg_heap::get_roots(vector<int>& scan, bool keep_identifiers) const
     insert_at_end(scan, stack); // inc_heads = yes
     insert_at_end(scan, temp); // yes
     insert_at_end(scan, heads); // yes
-    insert_at_end(scan, probability_heads); // yes
+    insert_at_end(scan, prior_heads); // yes
+    insert_at_end(scan, likelihood_heads); // yes
     insert_at_end(scan, random_modifiables_); // yes
     insert_at_end(scan, transition_kernels_); // yes
 
@@ -863,29 +1137,42 @@ int reg_heap::set_head(int index, int R2)
     return R1;
 }
 
-int reg_heap::allocate_head()
+int reg_heap::set_head(int index, closure&& C)
+{
+    int R = allocate();
+
+    set_head(index, R);
+
+    set_C(R, std::move(C) );
+
+    return R;
+}
+
+int reg_heap::allocate_head(closure&& C)
 {
     int R = allocate();
-    total_reg_allocations++;
 
     heads.push_back(R);
 
+    set_C(R, std::move(C));
+
     return R;
 }
 
 int reg_heap::push_temp_head()
 {
     int R = allocate();
-    total_reg_allocations++;
 
     temp.push_back(R);
 
     return R;
 }
 
-int reg_heap::push_temp_head(int R)
+int reg_heap::push_temp_head(closure&& C)
 {
-    temp.push_back(R);
+    int R = push_temp_head();
+
+    set_C(R, std::move(C));
 
     return R;
 }
@@ -897,51 +1184,43 @@ void reg_heap::pop_temp_head()
     temp.pop_back();
 }
 
-void reg_heap::get_more_memory()
+void reg_heap::resize(int s)
 {
-    collect_garbage();
-    base_pool_t::get_more_memory();
-}
+    assert(regs.size() == s);
 
-void reg_heap::expand_memory(int s)
-{
-    int old_size = size();
-    for(int t=0;t<tokens.size();t++)
+    auto old_size = prog_steps.size();
+    // Extend program.  Use regs.size() instead of size()
+    prog_steps.resize(regs.size());
+    prog_results.resize(regs.size());
+    prog_temp.resize(regs.size());
+
+    // Now we can use size() again.
+    for(auto i=old_size;i<size();i++)
     {
-	assert(prog_steps.size() == old_size);
-	assert(prog_results.size() == old_size);
-	assert(prog_temp.size() == old_size);
-    }
+	prog_steps[i] = non_computed_index;
+	prog_results[i] = non_computed_index;
 
-    base_pool_t::expand_memory(s);
-
-    // Extend program
-    prog_steps.resize(size());
-    prog_results.resize(size());
-    prog_temp.resize(size());
-    for(int i=old_size;i<size();i++)
-    {
-	assert(prog_steps[i] == 0);
-	assert(prog_results[i] == 0);
-	assert(prog_temp[i] == 0);
+	assert(prog_steps[i] == non_computed_index);
+	assert(prog_results[i] == non_computed_index);
+	assert(prog_temp[i].none());
     }
 }
 
 bool reg_heap::reg_is_constant(int r) const
 {
-    return access(r).type == reg::type_t::constant;
+    return regs.access(r).type == reg::type_t::constant;
 }
 
 bool reg_heap::reg_is_changeable(int r) const
 {
-    return access(r).type == reg::type_t::changeable;
+    return regs.access(r).type == reg::type_t::changeable;
 }
 
 void reg_heap::make_reg_changeable(int r)
 {
-    assert( access(r).type == reg::type_t::changeable or access(r).type == reg::type_t::unknown );
+    assert( regs.access(r).type == reg::type_t::changeable or regs.access(r).type == reg::type_t::unknown );
 
-    access(r).type = reg::type_t::changeable;
+    regs.access(r).type = reg::type_t::changeable;
 }
 
 bool reg_heap::result_is_called_by(int rc1, int rc2) const
@@ -955,8 +1234,8 @@ bool reg_heap::result_is_called_by(int rc1, int rc2) const
 
 bool reg_heap::result_is_used_by(int s1, int rc2) const
 {
-    for(int s: results[rc2].used_by)
-	if (s == s1)
+    for(auto& s: results[rc2].used_by)
+	if (s.first == s1)
 	    return true;
 
     return false;
@@ -1000,11 +1279,11 @@ void reg_heap::check_used_regs_in_token(int t) const
     for(auto p: tokens[t].delta_result())
     {
 	int r = p.first;
-	assert((prog_temp[r]&1) == 0);
-	prog_temp[r] |= 1;
-	if (is_root_token(t)) assert(p.second != -1);
+	assert(not prog_temp[r].test(0));
+	prog_temp[r].set(0);
+
 	// No results for constant regs
-	assert(access(r).type != reg::type_t::constant);
+	assert(regs.access(r).type != reg::type_t::constant);
 	int rc = p.second;
 	if (rc > 0)
 	{
@@ -1017,13 +1296,13 @@ void reg_heap::check_used_regs_in_token(int t) const
     for(auto p: tokens[t].delta_step())
     {
 	int r = p.first;
-	assert((prog_temp[r]&2) == 0);
-	prog_temp[r] |= 2;
-	if (is_root_token(t)) assert(p.second != -1);
+	assert(not prog_temp[r].test(1));
+	prog_temp[r].set(1);
+
 	// If the step is unshared, the result must be unshared as well: this allows us to just walk unshared results.
-	assert(prog_temp[r] == 3);
+	assert(prog_temp[r].test(0) and prog_temp[r].test(1));
 	// No steps for constant regs
-	assert(access(r).type != reg::type_t::constant);
+	assert(regs.access(r).type != reg::type_t::constant);
     }
 
     // FIXME - nonlocal. The same result/step are not set in multiple places!
@@ -1031,7 +1310,7 @@ void reg_heap::check_used_regs_in_token(int t) const
     for(auto p: tokens[t].delta_step())
     {
 	int r_s = p.second;
-	if (r_s <= 0) continue;
+	if (r_s < 0) continue;
 	
 	for(const auto& rcp2: steps[r_s].used_inputs)
 	{
@@ -1046,7 +1325,7 @@ void reg_heap::check_used_regs_in_token(int t) const
 
 	    // The used result should be referenced somewhere more root-ward
 	    // so that this result can be invalidated, and the used result won't be GC-ed.
-            // FIXME - nonlocal.  assert(is_modifiable(access(R2).C.exp) or result_is_referenced(t,rc2));
+            // FIXME - nonlocal.  assert(is_modifiable(regs.access(R2).C.exp) or result_is_referenced(t,rc2));
       
 	    // Used results should have values
 	    assert(results[rc2].value);
@@ -1057,7 +1336,7 @@ void reg_heap::check_used_regs_in_token(int t) const
     {
 	int r = p.first;
 	int r_r = p.second;
-	if (r_r <= 0) continue;
+	if (r_r < 0) continue;
 
 	int r_s = results[r_r].source_step;
 	int call = steps[r_s].call;
@@ -1069,19 +1348,22 @@ void reg_heap::check_used_regs_in_token(int t) const
 	if (results[r_r].flags.test(0))
 	    assert(is_root_token(t));
 
+	if (results[r_r].flags.test(1))
+	    assert(is_root_token(t));
+
 	if (value)
 	    assert(call);
 
 	if (call and value == call)
-	    assert(access(call).type == reg::type_t::constant);
+	    assert(regs.access(call).type == reg::type_t::constant);
 
-	if (call and value and access(call).type == reg::type_t::constant)
+	if (call and value and regs.access(call).type == reg::type_t::constant)
 	    assert(value == call);
 
 	if (t != root_token) continue;
 
 	// Regs with values should have back-references from their call.
-	if (value and access(call).type != reg::type_t::constant)
+	if (value and regs.access(call).type != reg::type_t::constant)
 	{
 	    assert( has_result(call) );
 	    int rc2 = result_index_for_reg(call);
@@ -1094,11 +1376,16 @@ void reg_heap::check_used_regs_in_token(int t) const
     }
 
     for(auto p: tokens[t].delta_result())
-	prog_temp[p.first] = 0;
+    {
+	prog_temp[p.first].reset(0);
+	prog_temp[p.first].reset(1);
+    }
 
     for(auto p: tokens[t].delta_step())
-	prog_temp[p.first] = 0;
-
+    {
+	prog_temp[p.first].reset(0);
+	prog_temp[p.first].reset(1);
+    }
 }
 
 void reg_heap::check_used_regs() const
@@ -1106,9 +1393,6 @@ void reg_heap::check_used_regs() const
     assert(tokens[root_token].vm_step.empty());
     assert(tokens[root_token].vm_result.empty());
 
-    for(auto c: prog_temp)
-	assert(not c);
-
     for(int t=0; t< tokens.size(); t++)
 	if (token_is_used(t))
 	    check_used_regs_in_token(t);
@@ -1121,9 +1405,6 @@ void reg_heap::check_used_regs() const
 	if (call > 0)
 	    assert(not results.is_free(call));
     }
-
-    for(auto c: prog_temp)
-	assert(not c);
 }
 
 int reg_heap::get_shared_step(int r)
@@ -1135,6 +1416,8 @@ int reg_heap::get_shared_step(int r)
     // 2. Set the source of the computation
     steps[s].source_reg = r;
 
+    assert(s > 0);
+    
     return s;
 }
 
@@ -1149,6 +1432,8 @@ int reg_heap::add_shared_step(int r)
     // Link it in to the mapping
     prog_steps[r] = s;
 
+    assert(s > 0);
+
     return s;
 }
 
@@ -1162,6 +1447,8 @@ int reg_heap::get_shared_result(int r, int s)
     results[rc].source_step = s;
     results[rc].source_reg = r;
 
+    assert(rc > 0);
+
     return rc;
 }
 
@@ -1178,49 +1465,86 @@ int reg_heap::add_shared_result(int r, int s)
     // Link it in to the mapping
     prog_results[r] = rc;
 
+    assert(rc > 0);
+
     return rc;
 }
 
 void reg_heap::check_back_edges_cleared_for_step(int s)
 {
     for(auto& rcp: steps.access_unused(s).used_inputs)
-	assert(null(rcp.second));
+	assert(rcp.second == 0);
     for(auto& r: steps.access_unused(s).created_regs)
     {
-	auto& created_by = access(r).created_by;
+	auto& created_by = regs.access(r).created_by;
 	assert(created_by.first == 0);
-	assert(null(created_by.second));
+	assert(created_by.second == 0);
     }
 }
 
 void reg_heap::check_back_edges_cleared_for_result(int rc)
 {
-    assert(null(results.access_unused(rc).call_edge.second));
+    assert(results.access_unused(rc).call_edge.second == 0);
 }
 
 void reg_heap::clear_back_edges_for_reg(int r)
 {
     assert(r > 0);
-    auto& created_by = access(r).created_by;
+    auto& created_by = regs.access(r).created_by;
     int s = created_by.first;
     if (s > 0)
     {
-	steps[s].created_regs.erase(created_by.second);
-	created_by.first = 0;
-	created_by.second = {};
+	auto& backward = steps[s].created_regs;
+	int j = created_by.second;
+	assert(0 <= j and j < backward.size());
+
+	// Clear the forward edge.
+	created_by = {0, 0};
+
+	// Move the last element to the hole, and adjust index of correspond forward edge.
+	if (j + 1 < backward.size())
+	{
+	    backward[j] = backward.back();
+	    auto& forward2 = regs.access(backward[j]);
+	    forward2.created_by.second = j;
+
+	    assert(regs.access(backward[j]).created_by.second == j);
+	}
+	backward.pop_back();
     }
 }
 
 void reg_heap::clear_back_edges_for_step(int s)
 {
     assert(s > 0);
-    for(auto& rcp: steps[s].used_inputs)
+    auto& forward1 = steps[s].used_inputs;
+    for(int i=0; i<forward1.size(); i++)
     {
-	results[rcp.first].used_by.erase(rcp.second);
-	rcp.second = {};
+	auto& backward = results[forward1[i].first].used_by;
+	assert(0 <= i and i < forward1.size());
+	int j = forward1[i].second;
+	assert(0 <= j and j < backward.size());
+
+	forward1[i] = {0,0};
+
+	if (j+1 < backward.size())
+	{
+	    // erase the backward edge by moving another backward edge on top of it.
+	    backward[j] = backward.back();
+	    // adjust the forward edge for that backward edge
+	    int i2 = backward[j].second;
+	    auto& forward2 = steps[backward[j].first].used_inputs;
+	    assert(0 <= i2 and i2 < forward2.size());
+	    forward2[i2].second = j;
+
+	    assert(steps[backward[j].first].used_inputs[backward[j].second].second == j);
+	    assert(results[forward2[i2].first].used_by[forward2[i2].second].second == i2);
+	}
+
+	backward.pop_back();
     }
     for(auto& r: steps[s].created_regs)
-	access(r).created_by = {0,{}};
+	regs.access(r).created_by = {0,{}};
     steps[s].created_regs.clear();
 }
 
@@ -1228,14 +1552,29 @@ void reg_heap::clear_back_edges_for_result(int rc)
 {
     assert(rc > 0);
     // FIXME! If there is a value, set, there should be a call_edge
-    // FIXME! Should we unmap all values with no .. value/call_edge?
+    //        Should we unmap all values with no .. value/call_edge?
     int call = results[rc].call_edge.first;
     if (call)
     {
 	assert(results[rc].value);
-	auto back_edge = results[rc].call_edge.second;
-	results[call].called_by.erase(back_edge);
-	results[rc].call_edge = {0,{}};
+
+	auto& backward = results[call].called_by;
+	int j = results[rc].call_edge.second;
+	assert(0 <= j and j < backward.size());
+
+	// Clear the forward edge.
+	results[rc].call_edge = {0, 0};
+
+	// Move the last element to the hole, and adjust index of correspond forward edge.
+	if (j+1 < backward.size())
+	{
+	    backward[j] = backward.back();
+	    auto& forward2 = results[backward[j]];
+	    forward2.call_edge.second = j;
+
+	    assert(results[backward[j]].call_edge.second == j);
+	}
+	backward.pop_back();
     }
 }
 
@@ -1243,7 +1582,7 @@ void reg_heap::clear_step(int r)
 {
     assert(not has_result(r));
     int s = prog_steps[r];
-    prog_steps[r] = 0;
+    prog_steps[r] = non_computed_index;
   
     if (s > 0)
     {
@@ -1257,7 +1596,7 @@ void reg_heap::clear_step(int r)
 void reg_heap::clear_result(int r)
 {
     int rc = prog_results[r];
-    prog_results[r] = 0;
+    prog_results[r] = non_computed_index;
 
     if (rc > 0)
     {
@@ -1278,7 +1617,7 @@ const expression_ref& reg_heap::get_parameter_value_in_context(int p, int c)
 const expression_ref& reg_heap::get_reg_value_in_context(int& R, int c)
 {
     total_get_reg_value++;
-    if (access(R).type == reg::type_t::constant) return access(R).C.exp;
+    if (regs.access(R).type == reg::type_t::constant) return regs.access(R).C.exp;
 
     total_get_reg_value_non_const++;
     reroot_at_context(c);
@@ -1287,7 +1626,7 @@ const expression_ref& reg_heap::get_reg_value_in_context(int& R, int c)
     {
 	total_get_reg_value_non_const_with_result++;
 	int R2 = result_value_for_reg(R);
-	if (R2) return access(R2).C.exp;
+	if (R2) return regs.access(R2).C.exp;
     }
 
     // If the value needs to be computed (e.g. its a call expression) then compute it.
@@ -1295,7 +1634,7 @@ const expression_ref& reg_heap::get_reg_value_in_context(int& R, int c)
     R = p.first;
     int value = p.second;
 
-    return access(value).C.exp;
+    return regs.access(value).C.exp;
 }
 
 void reg_heap::set_reg_value_in_context(int P, closure&& C, int c)
@@ -1328,7 +1667,7 @@ const closure& reg_heap::lazy_evaluate(int& R)
     auto p = incremental_evaluate(R);
     R = p.first;
     int value = p.second;
-    return access(value).C;
+    return regs.access(value).C;
 }
 
 const closure& reg_heap::lazy_evaluate(int& R, int c)
@@ -1336,7 +1675,7 @@ const closure& reg_heap::lazy_evaluate(int& R, int c)
     auto p = incremental_evaluate_in_context(R,c);
     R = p.first;
     int value = p.second;
-    return access(value).C;
+    return regs.access(value).C;
 }
 
 const closure& reg_heap::lazy_evaluate_head(int index, int c)
@@ -1348,18 +1687,18 @@ const closure& reg_heap::lazy_evaluate_head(int index, int c)
     if (R2 != R1)
 	set_head(index, R2);
 
-    return access(value).C;
+    return regs.access(value).C;
 }
 
 const closure& reg_heap::lazy_evaluate_unchangeable(int& R)
 {
     R = incremental_evaluate_unchangeable(R);
-    return access(R).C;
+    return regs.access(R).C;
 }
 
 int reg_heap::get_modifiable_value_in_context(int R, int c)
 {
-    assert( access(R).C.exp.head().type() == modifiable_type);
+    assert( regs.access(R).C.exp.head().type() == modifiable_type);
     assert( reg_is_changeable(R) );
 
     reroot_at_context(c);
@@ -1374,21 +1713,20 @@ int reg_heap::add_identifier(const string& name)
 	throw myexception()<<"Cannot add identifier '"<<name<<"': there is already an identifier with that name.";
 
     int R = allocate();
-    total_reg_allocations++;
 
     identifiers[name] = R;
     return R;
 }
 
 reg_heap::reg_heap(const std::shared_ptr<module_loader>& L)
-    :base_pool_t(1),
+    :regs(1,[this](int s){resize(s);}, [this](){collect_garbage();} ),
      steps(1),
      results(1),
      P(new Program(L)),
-     prog_steps(1),
-     prog_results(1),
+     prog_steps(1,non_existant_index),
+     prog_results(1, non_existant_index),
      prog_temp(1)
-{ 
+{
 }
 
 void reg_heap::release_scratch_list() const
