@@ -6,7 +6,7 @@
 #include "operations.H"
 #include "computation/machine/graph_register.H" // for graph_normalize( )
 #include "computation/operations.H"
-#include "parser/parse.H"
+#include "parser/rename.H"
 #include "parser/desugar.H"
 #include "computation/loader.H"
 #include "expression/AST_node.H"
@@ -20,6 +20,7 @@
 #include "computation/optimization/simplifier.H"
 #include "computation/optimization/occurrence.H"
 #include "computation/optimization/let-float.H"
+#include "computation/optimization/inliner.H"
 
 using std::pair;
 using std::map;
@@ -44,7 +45,7 @@ bool operator==(const symbol_info&S1, const symbol_info& S2)
 {
     return (S1.name == S2.name) and (S1.symbol_type == S2.symbol_type) and 
 	(S1.arity == S2.arity) and (S1.precedence == S2.precedence) and (S1.fixity == S2.fixity) and
-	(S1.type.ptr() == S2.type.ptr());
+	(S1.type == S2.type);
 }
 
 bool operator!=(const symbol_info&S1, const symbol_info& S2)
@@ -54,15 +55,6 @@ bool operator!=(const symbol_info&S1, const symbol_info& S2)
 
 symbol_info lookup_symbol(const string& name, const Program& P);
 
-bool Module::symbol_exists(const string& name) const
-{
-    auto loc = symbols.find(name);
-    if (loc == symbols.end())
-	return false;
-    else
-	return true;
-}
-
 void Module::add_symbol(const symbol_info& S)
 {
     if (is_haskell_builtin_con_name(S.name))
@@ -80,17 +72,15 @@ void Module::add_symbol(const symbol_info& S)
 
 void Module::add_alias(const string& identifier_name, const string& resolved_name)
 {
-    if (not symbol_exists(resolved_name))
+    if (not symbols.count(resolved_name))
 	throw myexception()<<"Can't add alias '"<<identifier_name<<"' -> '"<<resolved_name<<"' in module '"<<name<<"' because '"<<resolved_name<<"' is neither declared nor imported.";
 
-    std::pair<string,string> element(identifier_name,resolved_name);
-
     auto range = aliases.equal_range(identifier_name);
     for(auto it = range.first; it != range.second; it++)
 	if (it->second == resolved_name)
 	    return;
 
-    aliases.insert( std::pair<string,string>(identifier_name, resolved_name) );
+    aliases.insert( {identifier_name, resolved_name} );
 }
 
 void Module::declare_symbol(const symbol_info& S)
@@ -101,7 +91,7 @@ void Module::declare_symbol(const symbol_info& S)
     symbol_info S2 = S;
     S2.name = name + "." + S.name;
 
-    if (symbol_exists(S2.name))
+    if (symbols.count(S2.name))
 	throw myexception()<<"Trying to declare '"<<S.name<<"' twice in module '"<<name<<"'";
 
     // Add the symbol first.
@@ -120,52 +110,23 @@ void Module::declare_fixity(const std::string& s, int precedence, fixity_t fixit
     if (is_qualified_symbol(s))
 	throw myexception()<<"Trying to declare fixity of qualified symbol '"<<s<<"'.  Use its unqualified name.";
 
-    string s2 = name + "." + s;
-
-    if (not symbol_exists(s2))
-	declare_symbol({s, unknown_symbol, -1, -1, unknown_fix, {}});
-
-    symbol_info& S = symbols.find(s2)->second;
     if (precedence < 0 or precedence > 9)
 	throw myexception()<<"Precedence level "<<precedence<<" not allowed.";
+
     if (fixity == unknown_fix)
 	throw myexception()<<"Cannot set fixity to unknown!";
 
+    string s2 = name + "." + s;
+
+    if (not symbols.count(s2))
+	declare_symbol({s, unknown_symbol, -1, -1, unknown_fix, {}});
+
+    symbol_info& S = symbols.find(s2)->second;
+
     S.precedence = precedence;
     S.fixity = fixity;
 }
 
-void Module::add_import(bool qualified, const string& modid)
-{
-    vector<expression_ref> sub;
-    if (qualified)
-	sub.push_back(String("qualified"));
-    sub.push_back(String(modid));
-  
-    add_impdecl(expression_ref{AST_node("impdecl"),sub});
-}
-
-void Module::add_import_as(bool qualified, const string& modid, const string& modid2)
-{
-    vector<expression_ref> sub;
-    if (qualified)
-	sub.push_back(String("qualified"));
-    sub.push_back(String(modid));
-    sub.push_back(String("as"));
-    sub.push_back(String(modid2));
-  
-    add_impdecl(expression_ref{AST_node("impdecl"),sub});
-}
-
-void Module::add_impdecl(const expression_ref& impdecl)
-{
-    vector<expression_ref> sub;
-    if (impdecls)
-	sub = impdecls.sub();
-    sub.push_back(impdecl);
-    impdecls = expression_ref{AST_node("impdecls"),sub};
-}
-
 // Question: what if we import m1.s, which depends on an unimported m2.s?
 void Module::import_symbol(const symbol_info& S, const string& modid, bool qualified)
 {
@@ -181,23 +142,38 @@ void Module::import_symbol(const symbol_info& S, const string& modid, bool quali
 	add_alias(get_unqualified_name(S.name), S.name);
 }
 
-void Module::import_module(const Module& M2, const string& modid, bool qualified)
+void Module::import_module(const Program& P, const module_import& I)
 {
+    auto& M2 = P.get_module(I.name);
+    auto& modid = I.as;
+    bool qualified = I.qualified;
     assert(modid != name);
 
     // Right now 'exports' only has functions, not data types or constructors
 
-    for(const auto& p: M2.exported_symbols())
-    {
-	const symbol_info& S = p.second;
+    auto& m2_exports = M2.exported_symbols();
 
-	import_symbol(S, modid, qualified);
-    }
-}
+    if (I.only)
+	for(const auto& s_name: I.symbols)
+	{
+	    if (not m2_exports.count(s_name))
+		throw myexception()<<"Module '"<<I.name<<"' has no symbol '"<<s_name<<"'";
 
-void Module::import_module(const Module& M2, bool qualified)
-{
-    import_module(M2, M2.name, qualified);
+	    const symbol_info& S = m2_exports.at(s_name);
+
+	    import_symbol(S, modid, qualified);
+	}
+    else
+	for(const auto& p: m2_exports)
+	{
+	    const symbol_info& S = p.second;
+
+	    auto unqualified_name = get_unqualified_name(S.name);
+
+	    if (I.hiding and I.symbols.count(unqualified_name)) continue;
+
+	    import_symbol(S, modid, qualified);
+	}
 }
 
 module_import parse_import(const expression_ref& impdecl)
@@ -205,16 +181,27 @@ module_import parse_import(const expression_ref& impdecl)
     module_import mi;
 
     int i=0;
-    mi.qualified = impdecl.sub()[0].as_<String>() == "qualified";
+    mi.qualified = is_AST(impdecl.sub()[0], "qualified");
     if (mi.qualified) i++;
 	    
     mi.name = impdecl.sub()[i++].as_<String>();
 	    
-    if (i < impdecl.size() and impdecl.sub()[i++].as_<String>() == "as")
-	mi.as = impdecl.sub()[i++].as_<String>();
+    if (i < impdecl.size() and is_AST(impdecl.sub()[i],"as"))
+	mi.as = impdecl.sub()[i++].as_<AST_node>().value;
     else
 	mi.as = mi.name;
 
+    if (i < impdecl.size() and (is_AST(impdecl.sub()[i], "only") or is_AST(impdecl.sub()[i],"hiding")))
+    {
+	mi.hiding = is_AST(impdecl.sub()[i],"hiding");
+	mi.only = is_AST(impdecl.sub()[i],"only");
+	assert(mi.hiding or mi.only);
+
+	auto& objects = impdecl.sub()[i++].sub();
+	for(auto& x: objects)
+	    mi.symbols.insert(x.as_<AST_node>().value);
+    }
+
     // only:   handle import qualified A as B (x, y)
     // hiding:  handle import qualified A as B hiding (x, y)
 
@@ -222,31 +209,32 @@ module_import parse_import(const expression_ref& impdecl)
     return mi;
 }
 
-map<string,module_import> Module::imports() const
+vector<module_import> Module::imports() const
 {
-    std::map<string, module_import> imports_map;
+    vector<module_import> imports_list;
 
+    bool seen_Prelude = false;
     if (impdecls)
 	for(const auto& impdecl:impdecls.sub())
 	{
 	    auto import = parse_import(impdecl);
-	    if (imports_map.count(import.name))
-		throw myexception()<<"Module '"<<name<<"': importing module '"<<import.name<<"' twice!";
-	    imports_map[import.name] = import;
+	    if (import.name == "Prelude")
+		seen_Prelude = true;
+	    imports_list.push_back(import);
 	}
 
     // Import the Prelude if it wasn't explicitly mentioned in the import list.
-    if (not imports_map.count("Prelude") and name != "Prelude")
-	imports_map["Prelude"] = module_import("Prelude");
+    if (not seen_Prelude and name != "Prelude" and not language_options.count("NoImplicitPrelude"))
+	imports_list.push_back( module_import("Prelude") );
 
-    return imports_map;
+    return imports_list;
 }
 
 set<string> Module::dependencies() const
 {
     set<string> modules;
     for(auto& mi: imports())
-	modules.insert(mi.second.name);
+	modules.insert(mi.name);
     return modules;
 }
 
@@ -255,15 +243,20 @@ void Module::compile(const Program& P)
     assert(not resolved);
     resolved = true;
 
-    add_local_symbols();
-
     perform_imports(P);
 
+    declare_fixities();
+
+    rename_infix(P);  // FIXME - merge with rename() below.
+
+    add_local_symbols();
+
     perform_exports();
 
     // Currently we do renaming here, including adding prefixes to top-level decls.
-    if (not skip_desugaring)
-	desugar(P); // fixme - separate renaming from desugaring -- move it after load_builtins.
+    rename(P);
+
+    desugar(P);
 
     load_builtins(*P.get_module_loader());
 
@@ -295,10 +288,7 @@ void Module::compile(const Program& P)
 void Module::perform_imports(const Program& P)
 {
     for(auto& i: imports())
-    {
-	auto& I = P.get_module(i.second.name);
-	import_module(I, i.second.as, i.second.qualified);
-    }
+	import_module(P, i);
 }
 
 void Module::export_symbol(const symbol_info& S)
@@ -314,7 +304,7 @@ void Module::export_symbol(const symbol_info& S)
 
 void Module::export_module(const string& modid)
 {
-    if (modid != name and not imports().count(modid))
+    if (modid != name and not dependencies().count(modid))
 	throw myexception()<<"module '"<<modid<<"' is exported but not imported";
 
     for(auto& s: symbols)
@@ -341,7 +331,7 @@ void Module::perform_exports()
 	{
 	    if (is_AST(ex,"qvar"))
 	    {
-		string qvarid = ex.as_<AST_node>().value;
+		string qvarid = ex.as_<AST_node>().value; // This ignores export subspec - see grammar.
 		if (aliases.count(qvarid))
 		    export_symbol(lookup_symbol(qvarid));
 		else
@@ -383,17 +373,37 @@ map<string,expression_ref> Module::code_defs() const
     return code;
 }
 
-void Module::desugar(const Program&)
+void Module::rename_infix(const Program&)
+{
+    if (topdecls)
+    {
+	assert(is_AST(topdecls,"TopDecls"));
+	topdecls = ::rename_infix(*this,topdecls);
+    }
+}
+
+void Module::rename(const Program& P)
+{
+    if (topdecls)
+    {
+	assert(is_AST(topdecls,"TopDecls"));
+	topdecls = ::rename(*this,topdecls);
+    }
+    if (P.get_module_loader()->dump_renamed)
+	std::cout<<name<<"[renamed]:\n"<<topdecls<<"\n\n";
+}
+
+void Module::desugar(const Program& P)
 {
     if (topdecls)
     {
 	assert(is_AST(topdecls,"TopDecls"));
 	topdecls = ::desugar(*this,topdecls);
     }
+    if (P.get_module_loader()->dump_desugared)
+	std::cout<<name<<"[desugared]:\n"<<topdecls<<"\n\n";
 }
 
-int nodes_size(const expression_ref& E);
-
 void add_constructor(map<var,expression_ref>& decls, const constructor& con)
 {
     var x(con.name());
@@ -435,21 +445,30 @@ void Module::export_small_decls()
     if (not topdecls) return;
 
     assert(small_decls_out.empty());
+
+    // Modules that we imported should have their small_decls transitively inherited
+    small_decls_out = small_decls_in;
+
     for(auto& decl: topdecls.sub())
     {
 	auto& x = decl.sub()[0].as_<var>();
 	assert(not x.name.empty());
 
 	auto& body = decl.sub()[1];
-	if (nodes_size(body) < 15)
+	if (simple_size(body) <= 5)
+	{
 	    small_decls_out.insert({x, body});
+	}
     }
 
+    // Find free vars in the decls that are not bound by *other* decls.
     for(auto& decl: small_decls_out)
     {
 	set<var> free_vars;
 	tie(decl.second, free_vars) = occurrence_analyzer(decl.second);
-	small_decls_out_free_vars.insert(free_vars.begin(), free_vars.end());
+	for(auto& x: free_vars)
+	    if (not small_decls_out.count(x))
+		small_decls_out_free_vars.insert(x);
     }
 }
 
@@ -701,17 +720,6 @@ vector<expression_ref> peel_lambdas(expression_ref& E)
     return args;
 }
 
-int nodes_size(const expression_ref& E)
-{
-    int total = 1;
-
-    if (E.is_expression())
-	for(const auto& e:E.sub())
-	    total += nodes_size(e);
-
-    return total;
-}
-
 void mark_exported_decls(CDecls& decls, const map<string,symbol_info>& exports, const string& module_name)
 {
     // Record exports
@@ -761,7 +769,8 @@ void Module::optimize(const Program& P)
 		// This won't float things to the top level!
 		auto name = decl.sub()[0].as_<var>().name;
 		auto body = decl.sub()[1];
-		body = let_float(body);
+		if (P.get_module_loader()->fully_lazy)
+		    body = let_float(body);
 		body = graph_normalize(body);
 
 		new_decls.push_back(AST_node("Decl") + decl.sub()[0] + body);
@@ -845,20 +854,15 @@ void Module::load_constructors()
 		assert(is_AST(constrs,"constrs"));
 		for(const auto& constr: constrs.sub())
 		{
-		    string cname;
-		    int arity = -1;
-		    if (is_AST(constr,"constr"))
+		    int arity = 0;
+		    expression_ref type = constr;
+		    if (is_AST(type,"TypeApply"))
 		    {
-			cname = constr.sub()[0].as_<String>();
-			arity = constr.size() - 1;
+			arity = type.size()-1;
+			type = type.sub()[0];
 		    }
-		    else if (is_AST(constr,"constr_op"))
-		    {
-			cname = constr.sub()[1].as_<String>();
-			arity = 2;
-		    }
-		    else
-			std::abort();
+		    assert(is_AST(type,"type_id"));
+		    string cname = type.head().as_<AST_node>().value;
 		    string qualified_name = name+"."+cname;
 		    expression_ref body = lambda_expression( constructor(qualified_name, arity) );
 		    new_decls.push_back(AST_node("Decl") + var(qualified_name) + body);
@@ -877,14 +881,6 @@ bool Module::is_declared(const std::string& name) const
     return is_haskell_builtin_con_name(name) or (aliases.count(name) > 0);
 }
 
-bool Module::is_declared_local(const std::string& sname) const
-{
-    auto loc = symbols.find(sname);
-    if (loc == symbols.end()) return false;
-
-    return (get_module_name(loc->second.name) == name);
-}
-
 pair<symbol_info,expression_ref> Module::lookup_builtin_symbol(const std::string& name)
 {
     if (name == "()")
@@ -913,7 +909,7 @@ symbol_info Module::lookup_symbol(const std::string& name) const
     else if (count == 1)
     {
 	string symbol_name = aliases.find(name)->second;
-	if (not symbol_exists(symbol_name))
+	if (not symbols.count(symbol_name))
 	    throw myexception()<<"Identifier '"<<name<<"' -> '"<<symbol_name<<"', which does not exist!";
 	return symbols.find(symbol_name)->second;
     }
@@ -928,6 +924,16 @@ symbol_info Module::lookup_symbol(const std::string& name) const
     }
 }
 
+symbol_info Module::lookup_resolved_symbol(const std::string& symbol_name) const
+{
+    if (is_haskell_builtin_con_name(symbol_name))
+	return lookup_builtin_symbol(symbol_name).first;
+
+    if (not symbols.count(symbol_name))
+	throw myexception()<<"Identifier '"<<name<<"' -> '"<<symbol_name<<"', which does not exist!";
+    return symbols.find(symbol_name)->second;
+}
+
 symbol_info Module::get_operator(const string& name) const
 {
     symbol_info S = lookup_symbol(name);
@@ -1132,30 +1138,6 @@ string get_unqualified_name(const std::string& s)
 // We can't determine function bodies at all, since we can't even handle op definitions
 //   before we know fixities!
 
-string get_function_name(const expression_ref& E)
-{
-    if (is_AST(E,"funlhs1"))
-    {
-	expression_ref f = E.sub()[0];
-	assert(is_AST(f,"id"));
-
-	return f.head().as_<AST_node>().value;
-    }
-    else if (is_AST(E,"funlhs2"))
-    {
-	expression_ref f = E.sub()[1];
-	assert(is_AST(f,"id"));
-	return f.head().as_<AST_node>().value;
-    }
-    else if (is_AST(E,"funlhs3"))
-	return get_function_name(E.sub()[0]);
-    std::abort();
-}
-
-set<string> find_bound_vars(const expression_ref& E);
-set<string> find_all_ids(const expression_ref& E);
-
-
 void Module::def_function(const std::string& fname)
 {
     if (is_qualified_symbol(fname))
@@ -1201,7 +1183,7 @@ void Module::def_constructor(const std::string& cname, int arity)
     declare_symbol( {cname, constructor_symbol, arity, -1, unknown_fix, {}} );
 }
 
-void Module::add_local_symbols()
+void Module::declare_fixities()
 {
     if (not topdecls) return;
 
@@ -1235,14 +1217,26 @@ void Module::add_local_symbols()
 		declare_fixity(name, precedence, fixity);
 	    }
 	}
-	else if (is_AST(decl,"Decl"))
+}
+
+void Module::add_local_symbols()
+{
+    if (not topdecls) return;
+
+    assert(is_AST(topdecls,"TopDecls"));
+
+    // 0. Get names that are being declared.
+    for(const auto& decl: topdecls.sub())
+	if (is_AST(decl,"Decl"))
 	{
-	    expression_ref lhs = decl.sub()[0];
 	    set<string> vars;
-	    if (is_AST(lhs,"funlhs1") or is_AST(lhs,"funlhs2") or is_AST(lhs,"funlhs3"))
-		vars.insert( get_function_name(lhs) );
+	    if (is_function_binding(decl))
+		vars.insert( get_func_name(decl) );
 	    else
+	    {
+		auto& lhs = decl.sub()[0];
 		vars = find_bound_vars(lhs);
+	    }
 
 	    for(const auto& var_name: vars)
 	    {
@@ -1271,23 +1265,18 @@ void Module::add_local_symbols()
 	    if (decl.size() >= 2)
 	    {
 		expression_ref constrs = decl.sub()[1];
-		assert(is_AST(constrs,"constrs"));
 		for(const auto& constr: constrs.sub())
 		{
-		    if (is_AST(constr,"constr"))
+		    int arity = 0;
+		    auto type = constr;
+		    if (is_AST(type,"TypeApply"))
 		    {
-			string name = constr.sub()[0].as_<String>();
-			int arity = constr.size() - 1;
-			def_constructor(name,arity);
+			arity = type.size()-1;
+			type = type.sub()[0];
 		    }
-		    else if (is_AST(constr,"constr_op"))
-		    {
-			string name = constr.sub()[1].as_<String>();
-			int arity = 2;
-			def_constructor(name,arity);
-		    }
-		    else
-			std::abort();
+		    assert(is_AST(type,"type_id"));
+		    string cname = type.head().as_<AST_node>().value;
+		    def_constructor(cname,arity);
 		}
 	    }
 	}
@@ -1305,7 +1294,8 @@ Module::Module(const char *n)
     :Module(string(n))
 { }
 
-Module::Module(const expression_ref& E)
+Module::Module(const expression_ref& E, const set<string>& lo)
+    :language_options(lo)
 {
     assert(is_AST(E,"Module"));
     assert(not module);
