@@ -63,6 +63,8 @@ bool is_trivial(const expression_ref& E)
 
 void bind_var(in_scope_set& bound_vars, const var& x, const expression_ref& E)
 {
+    assert(x.index >= 0);
+    assert(not is_wildcard(E));
     assert(not bound_vars.count(x));
     assert(x.work_dup != amount_t::Unknown);
     assert(x.code_dup != amount_t::Unknown);
@@ -401,13 +403,24 @@ expression_ref rebuild_case(const simplifier_options& options, const expression_
 
 	// 3. If we know something extra about the value (or range, theoretically) of the object in this case branch, then record that.
 	bound_variable_info original_binding;
-	if (is_var(object)) original_binding = rebind_var(bound_vars, object.as_<var>(), patterns[i]);
+	if (is_var(object))
+	{
+	    if (is_wildcard(patterns[i]))
+		; // we know that the object does NOT match any of the patterns on the other branches
+	          // we could record a negative range.
+	    else
+		original_binding = rebind_var(bound_vars, object.as_<var>(), patterns[i]);
+	}
 
 	// 4. Simplify the alternative body
 	bodies[i] = simplify(options, bodies[i], S2, bound_vars, unknown_context());
 
 	// 5. Restore informatation about an object variable to information outside this case branch.
-	if (is_var(object)) rebind_var(bound_vars, object.as_<var>(), original_binding.first);
+	if (is_var(object))
+	{
+	    if (not is_wildcard(patterns[i]))
+		rebind_var(bound_vars, object.as_<var>(), original_binding.first);
+	}
 
 	unbind_decls(bound_vars, pat_decls);
     }
@@ -596,11 +609,33 @@ simplify_decls(const simplifier_options& options, CDecls& orig_decls, const subs
 	}
 	else
 	{
+	    /* See the paper on inlining for why we should NOT consider if F is trivial until after we have
+                 simplified F.
+
+	       One case that seems kind of problematic is let {x=3,y=4,q=let {z=x,y=w} in (z,w)} in ...
+               If we simplify the bodies of z and w then we get let {x=3,y=4,q=let {z=3,w=4} in (z,w)} in ...
+                 thus (a) duplicating the bodies of x and y, and also making q not look like a constructor.
+               This is kind of problematic since we will never be able to resolve `case q of (a,b) -> a`
+                 unless we can expose that q is in fact a pair tuple.
+
+               However, consider that we might have decided to unconditionally pre-inline x.  In that case,
+                 replacing z with x is bad, since we would then replace z with 3 and get (3,w), which is illegal.
+               This the warning about not replacing z with x until after we simplify F (below) applies to this case.
+
+               In practice, it seems that we get around this problem by floating the let {z=w,y=w} up since (z,w)
+                 is a constructor.  However, if we could somehow avoid substituting x unconditionally in these cases,
+                 then we could indeed replace z with x.  Maybe we could avoid marking x as OnceSafe if its only occurrence
+                 is in the body of a let?  But then we would have to make sure that we never CREATE the situation where x
+                 is the body of a let.  So, maybe we just hope that floating lets out of less is sufficient for now.
+	    */
+
 	    // 5.1.2 Simplify F.
 	    F = simplify(options, F, S2, bound_vars, unknown_context());
 
+	    // Should we also float lambdas in addition to constructors?  We could apply them if so...
+
 	    // Float lets out of decl x = F
-	    if (options.let_float_from_let and (is_constructor(multi_let_body(F)) or is_top_level))
+	    if (options.let_float_from_let and (is_constructor_exp(multi_let_body(F)) or is_lambda_exp(multi_let_body(F)) or is_top_level))
 		for(auto& decls: strip_multi_let(F))
 		    for(auto& decl: decls)
 		    {
@@ -635,14 +670,14 @@ simplify_decls(const simplifier_options& options, CDecls& orig_decls, const subs
 //                     \x2 ->               ($) f x2  ===>              f
 //                 We don't do this (could perform more allocation):
 //                     \x2 -> (let decls in ($) f x2) ===> let decls in f
-expression_ref eta_reduce(const expression_ref& E)
+expression_ref maybe_eta_reduce(const expression_ref& E)
 {
-    assert(is_lambda(E.head()));
+    assert(is_lambda_exp(E));
     auto& x    = E.sub()[0].as_<var>();
     auto& body = E.sub()[1];
 
     if (x.code_dup == amount_t::Once and
-	body.is_expression() and is_apply(body.head()) and
+	body.is_expression() and is_apply_exp(body) and
 	(body.as_expression().sub.back() == x))
     {
 	// ($) f x  ==> f
@@ -658,7 +693,7 @@ expression_ref eta_reduce(const expression_ref& E)
 	}
     }
     else
-	return E;
+	return {};
 }
 
 // Q1. Where do we handle beta-reduction (@ constant x1 x2 ... xn)?
@@ -699,14 +734,22 @@ expression_ref simplify(const simplifier_options& options, const expression_ref&
     if (not E.size()) return E;
 
     // 2. Lambda (E = \x -> body)
-    if (is_lambda(E.head()))
+    if (is_lambda_exp(E))
     {
 	assert(E.size() == 2);
 
+	// 2.1 eta reduction: (\x -> @ f a1 ...... an x) => (@ f a1 ....... an)
+	if (auto E2 = maybe_eta_reduce(E))
+	{
+	    // Simplifying the body in (\x -> let {y=x} in f y y) can result in f x x even if x is originally only used once.
+	    // Since maybe_eta_reduce( ) uses occurrence info to check that x is only used once, we have to do this *before* we simplify E (below).
+	    return simplify(options, E2, S, bound_vars, context);
+	}
+
 	auto S2 = S;
 
 	auto Evar = E.sub()[0];
-	// 2.1. Get the new name, possibly adding a substitution
+	// 2.2. Get the new name, possibly adding a substitution
 	var x2 = rename_and_bind_var(Evar, S2, bound_vars);
 
 	// 2.3 Simplify the body with x added to the bound set.
@@ -716,7 +759,7 @@ expression_ref simplify(const simplifier_options& options, const expression_ref&
 	unbind_var(bound_vars,x2);
 
 	// 2.5 Return (\x2 -> new_body) after eta-reduction
-	return eta_reduce(lambda_quantify(x2, new_body));
+	return lambda_quantify(x2, new_body);
     }
 
     // 6. Case
@@ -733,7 +776,7 @@ expression_ref simplify(const simplifier_options& options, const expression_ref&
     }
 
     // ?. Apply
-    if (is_apply(E.head()))
+    if (is_apply_exp(E))
     {
 	object_ptr<expression> V2 = E.as_expression().clone();
 	
@@ -751,7 +794,7 @@ expression_ref simplify(const simplifier_options& options, const expression_ref&
     }
 
     // 4. Constructor or Operation
-    if (is_constructor(E.head()) or is_non_apply_operation(E.head()))
+    if (is_constructor_exp(E) or is_non_apply_op_exp(E))
     {
 	object_ptr<expression> E2 = E.as_expression().clone();
 	for(int i=0;i<E.size();i++)
