@@ -55,7 +55,6 @@ void add_modifiable_MH_move(const Model& P, const string& name, const proposal_f
 			    MCMC::MoveAll& M, double weight=1)
 {
     double rate = P.get_rate_for_reg(r);
-    std::cerr<<"rate = "<<rate<<"\n";
     M.add(rate * weight, MCMC::MH_Move( Proposal2M(proposal, r, parameters), name) );
 }
 
@@ -113,7 +112,6 @@ double default_sampling_rate(const Model& /*M*/, const string& /*parameter_name*
 bool add_slice_move(const Model& P, int r, MCMC::MoveAll& M, double weight = 1.0)
 {
     double rate = P.get_rate_for_reg(r);
-    std::cerr<<"rate = "<<rate<<"\n";
     auto range = P.get_range_for_reg(r);
     if (not range.is_a<Bounds<double>>()) return false;
 
@@ -174,7 +172,7 @@ void add_integer_uniform_MH_moves(const Model& P, MCMC::MoveAll& M, double weigh
 	string name = "m_int_uniform_"+convertToString<int>(r);
 	double l = (int)bounds.lower_bound;
 	double u = (int)bounds.upper_bound;
-	add_modifiable_MH_move(P,name, discrete_uniform, r, {double(l),double(u)}, M, weight);
+	add_modifiable_MH_move(P,name, discrete_uniform_avoid, r, {double(l),double(u)}, M, weight);
     }
 }
 
@@ -184,7 +182,6 @@ void add_integer_slice_moves(const Model& P, MCMC::MoveAll& M, double weight)
     {
 	auto range = P.get_range_for_reg(r);
 	double rate = P.get_rate_for_reg(r);
-	std::cerr<<"rate = "<<rate<<"\n";
 	if (not range.is_a<Bounds<int>>()) continue;
 
 	// FIXME: righteousness.
@@ -243,6 +240,22 @@ void add_alignment_and_parameter_moves(MCMC::MoveAll& moves, Model& M, double we
 	    moves.add(weight, MCMC::MH_Move(Generic_Proposal(proposal),"realign_and_sample_"+pname2), enabled);
 	}
     }
+
+    int n_scales = dynamic_cast<const Parameters&>(M).n_branch_scales();
+    for(int s=0; s<n_scales; s++)
+    {
+	string pname = "Scale["+std::to_string(s+1)+"]";
+	vector<int> partitions = dynamic_cast<const Parameters&>(M).partitions_for_scale(s);
+
+	if (auto r = dynamic_cast<const Parameters&>(M).branch_scale_modifiable_reg(s))
+	{
+	    auto proposal = [r,partitions](Model& P){
+		return realign_and_propose_parameter(P, *r, partitions, log_scaled(more_than(0.0, shift_laplace)), {0.25}) ;
+	    };
+
+	    moves.add(weight, MCMC::MH_Move(Generic_Proposal(proposal),"realign_and_sample_"+pname), enabled);
+	}
+    }
 }
 
 //FIXME - how to make a number of variants with certain things fixed, for burn-in?
@@ -545,7 +558,7 @@ MCMC::MoveAll get_parameter_MH_but_no_slice_moves(Model& M)
     // Also, how hard would it be to make a Gibbs flipper?  We could (perhaps) run that once per iteration to avoid periodicity.
 
     add_boolean_MH_moves(M, parameter_moves, 1.5);
-    add_integer_uniform_MH_moves(M, parameter_moves, 0.1);
+    add_integer_uniform_MH_moves(M, parameter_moves, 1.0);
 
     // Actually there ARE slice moves for this, but they don't jump modes!
     add_real_MH_moves(M, parameter_moves, 0.1);
@@ -685,7 +698,7 @@ void do_pre_burnin(const variables_map& args, owned_ptr<Model>& P, ostream& out_
     MoveStats Stats;
 
     // 0. Then sample  (a) scale and (b) branch lengths and (c) parameters
-    if (P->contains_key("pre-burnin-A"))
+    if (P->contains_key("pre-burnin-A") or P->contains_key("pre-burnin-A-pre"))
     {
 	MoveAll pre_burnin("pre-burnin");
 
@@ -705,14 +718,14 @@ void do_pre_burnin(const variables_map& args, owned_ptr<Model>& P, ostream& out_
 	    // turn training on
 	    {
 		Parameters& PP = *P.as<Parameters>();
-		PP.set_parameter_value(PP.find_parameter("*IModels.training"), new constructor("Prelude.True",0));
+		PP.set_parameter_value(PP.find_parameter("*IModels.training"), bool_true);
 	    }
 
 	    pre_burnin.iterate(P,Stats);
 	    // turn training off
 	    {
 		Parameters& PP = *P.as<Parameters>();
-		PP.set_parameter_value(PP.find_parameter("*IModels.training"), new constructor("Prelude.False",0));
+		PP.set_parameter_value(PP.find_parameter("*IModels.training"), bool_false);
 	    }
 
 	    log_preburnin(out_both, *P, "(S)+(L)+(P)+Alignment", i);
@@ -852,20 +865,23 @@ void do_pre_burnin(const variables_map& args, owned_ptr<Model>& P, ostream& out_
 
 
     // 4. Then do an initial tree search - SPR - with variable alignment
-    if (P->contains_key("pre-burnin-A"))
+    if (P->contains_key("pre-burnin-A") or P->contains_key("pre-burnin-A-post"))
     {
 	// turn training on
 	{
 	    Parameters& PP = *P.as<Parameters>();
-	    PP.set_parameter_value(PP.find_parameter("*IModels.training"), new constructor("Prelude.True",0));
+	    PP.set_parameter_value(PP.find_parameter("*IModels.training"), bool_true);
 	}
 
 	MoveAll pre_burnin("pre-burnin+A");
 	pre_burnin.add(4,get_scale_slice_moves(*P.as<Parameters>()));
 	if (all_scales_modifiable(*P))
 	    pre_burnin.add(4,MCMC::SingleMove(scale_means_only, "scale_means_only", "mean"));
-	pre_burnin.add(1,SingleMove(walk_tree_sample_branch_lengths, "walk_tree_sample_branch_lengths", "tree:lengths"));
-	pre_burnin.add(1,SingleMove(sample_SPR_A_search_all,"SPR_search_all", "tree:topology:lengths"));
+	pre_burnin.add(2, SingleMove(walk_tree_sample_NNI_and_branch_lengths, "NNI_and_lengths","tree:topology:lengths"));
+	// FIXME - add a fixed-topology version of realign_from_tips
+	pre_burnin.add(1, SingleMove(realign_from_tips, "realign_from_tips","lengths:alignment:topology") );
+	pre_burnin.add(2, get_parameter_slice_moves(*P));
+//	pre_burnin.add(1,SingleMove(sample_SPR_A_search_all,"SPR_search_all", "tree:topology:lengths"));
 
 	// enable and disable moves
 	enable_disable_transition_kernels(pre_burnin,args);
@@ -880,7 +896,7 @@ void do_pre_burnin(const variables_map& args, owned_ptr<Model>& P, ostream& out_
 	// turn training back off
 	{
 	    Parameters& PP = *P.as<Parameters>();
-	    PP.set_parameter_value(PP.find_parameter("*IModels.training"), new constructor("Prelude.False",0));
+	    PP.set_parameter_value(PP.find_parameter("*IModels.training"), bool_false);
 	}
 	for(int i=0;i<n_pre_burnin;i++)
 	{
@@ -942,7 +958,7 @@ void do_sampling(const variables_map& args,
 	//-------------- parameters (parameters_moves) --------------//
 	if (PP->variable_alignment() and PP->t().n_nodes() > 1)
 	{
-	    double factor = P->load_value("alignment_sampling_factor",1.0);
+	    double factor = P->load_value("alignment-sampling-factor",1.0);
 	    std::cerr<<"alignment sampling factor = "<<factor<<"\n";
 	    sampler.add(factor,alignment_moves);
 	}
