@@ -39,6 +39,8 @@
 #include "probability/probability.H"
 #include "computation/expression/expression.H"
 #include "computation/expression/lambda.H"
+#include "computation/expression/bool.H"
+#include "computation/expression/case.H"
 #include "computation/expression/tuple.H"
 #include "computation/expression/list.H"
 #include "computation/expression/var.H"
@@ -240,7 +242,7 @@ int data_partition::n_states() const
 vector<double> data_partition::distribution() const
 {
     int s = *P->smodel_index_for_partition(partition_index);
-    return P->evaluate(P->PC->SModels[s].distribution).as_<Vector<double>>();
+    return P->evaluate(P->PC->SModels[s].distribution).as_<EVector>();
 }
 
 Matrix data_partition::WeightedFrequencyMatrix() const
@@ -342,11 +344,6 @@ const pairwise_alignment_t& data_partition::get_pairwise_alignment(int b) const
     return get_pairwise_alignment_(b).as_<pairwise_alignment_t>();
 }
 
-log_double_t data_partition::prior_no_alignment() const 
-{
-    return 1.0;
-}
-
 // We want to decrease 
 // (a) the number of times get_counts( ) is called
 // (b) the number of times seqlength( ) is called
@@ -362,11 +359,6 @@ log_double_t data_partition::prior_alignment() const
     return Pr;
 }
 
-log_double_t data_partition::prior() const 
-{
-    return prior_alignment() * prior_no_alignment();
-}
-
 const Likelihood_Cache_Branch& data_partition::cache(int b) const
 {
     return P->evaluate(DPC().conditional_likelihoods_for_branch[b]).as_<Likelihood_Cache_Branch>();
@@ -651,7 +643,7 @@ data_partition_constants::data_partition_constants(Parameters* p, int i, const a
 	for(int n=0;n<sequence_length_pr_indices.size();n++)
 	{
 	    expression_ref l = p->get_expression(sequence_length_indices[n]);
-	    expression_ref lengthp = {var("Prelude.snd"),model};
+	    expression_ref lengthp = {var("Data.Tuple.snd"),model};
 	    sequence_length_pr_indices[n] = p->add_compute_expression( {lengthp,l} );
 	}
 
@@ -671,13 +663,19 @@ data_partition_constants::data_partition_constants(Parameters* p, int i, const a
 	}
 	else
 	    alignment_prior_index = p->add_compute_expression( {var("Alignment.alignment_pr"), as, p->my_tree(), hmms, model} );
+
+	p->add_prior_factor(make_if_expression(parameter("*variable_alignment"),
+					       p->get_expression(alignment_prior_index),
+					       log_double_t(1.0)));
     }
+
+    p->add_likelihood_factor(p->get_expression(likelihood_index));
 }
 
 //-----------------------------------------------------------------------------//
 smodel_methods::smodel_methods(const expression_ref& E, context& C)
 {
-    expression_ref V = var("Prelude.listToVectorDouble");
+    expression_ref V = var("Prelude.list_to_vector");
 
     main = C.add_compute_expression( E );
     expression_ref S = C.get_expression(main);
@@ -791,7 +789,7 @@ tree_constants::tree_constants(Parameters* p, const SequenceTree& T, const model
     {
 	string prefix = "T:lengths";
 	expression_ref branch_lengths = {branch_length_model.expression, tree};
-	branch_lengths = {var("Distributions.run_random'"), var("Prelude.Nothing"), 0.0, branch_lengths};
+	branch_lengths = {var("Distributions.gen_model_no_alphabet"), branch_lengths};
 	branch_lengths = {var("Distributions.do_log"), prefix, branch_lengths};
 	branch_lengths = {var("Prelude.unsafePerformIO"),branch_lengths};
 	branch_lengths = {var("Parameters.evaluate"),-1,branch_lengths};
@@ -824,8 +822,15 @@ tree_constants::tree_constants(Parameters* p, const SequenceTree& T, const model
 	std::cerr<<"Warning!  Some branch lengths not set because they are not directly modifiable.\n\n";
 }
 
+bool Parameters::variable_alignment_from_param() const
+{
+    auto e = get_parameter_value(variable_alignment_param);
+    return is_bool_true(e);
+}
+
 bool Parameters::variable_alignment() const
 {
+    assert(variable_alignment_from_param() == variable_alignment_);
     return variable_alignment_;
 }
 
@@ -837,6 +842,9 @@ void Parameters::variable_alignment(bool b)
     if (not n_imodels())
 	variable_alignment_ = false;
 
+    if (variable_alignment_from_param() != variable_alignment_)
+	set_parameter_value(variable_alignment_param, variable_alignment_);
+
     // turning ON alignment variation
     if (variable_alignment())
 	assert(n_imodels() > 0);
@@ -1157,6 +1165,7 @@ void Parameters::NNI(int b1, int b2, bool allow_disconnect_subtree)
 	dp.set_pairwise_alignment(t().find_branch(nodes[3],nodes[5]), get_pairwise_alignment_from_bits(a123456[i], 3, 5));
 	dp.set_pairwise_alignment(t().find_branch(nodes[4],nodes[5]), get_pairwise_alignment_from_bits(a123456[i], 4, 5));
     }
+
 }
 
 void Parameters::show_h_tree() const
@@ -1172,29 +1181,14 @@ void Parameters::show_h_tree() const
 
 log_double_t Parameters::prior_no_alignment() const 
 {
-    log_double_t Pr = Model::prior();
-
-    // prior for each branch being aligned/unaliged
-    if (variable_alignment()) 
+    if (variable_alignment_)
     {
-	const double p_unaligned = load_value("P_aligned",0.0);
-
-	log_double_t pNA = p_unaligned;
-
-	log_double_t pA = (1.0 - p_unaligned);
-
-	for(int b=0;b<t().n_branches();b++)
-	    if (not branch_HMM_type(b))
-		Pr *= pA;
-	    else
-		Pr *= pNA;
+	auto P2 = *this;
+	P2.variable_alignment(false);
+	return P2.prior();
     }
-
-    // prior on parameters in data partitions
-    for(int i=0;i<n_data_partitions();i++)
-	Pr *= get_data_partition(i).prior_no_alignment();
-
-    return Pr;
+    else
+	return prior();
 }
 
 log_double_t Parameters::prior_alignment() const 
@@ -1207,29 +1201,6 @@ log_double_t Parameters::prior_alignment() const
     return Pr;
 }
 
-log_double_t Parameters::prior() const 
-{
-    return prior_no_alignment() * prior_alignment();
-}
-
-log_double_t Parameters::likelihood() const 
-{
-    log_double_t Pr = 1;
-    for(int i=0;i<n_data_partitions();i++) 
-	Pr *= get_data_partition(i).likelihood();
-    return Pr;
-}
-
-log_double_t Parameters::heated_likelihood() const 
-{
-    log_double_t Pr = 1;
-
-    for(int i=0;i<n_data_partitions();i++) 
-	Pr *= get_data_partition(i).heated_likelihood();
-
-    return Pr;
-}
-
 void Parameters::select_root(int b) const
 {
     if (t().source(b) == subst_root() or t().target(b) == subst_root())
@@ -1289,12 +1260,14 @@ int Parameters::branch_scale_index(int i) const
     return PC->scale_parameter_indices[i];
 }
 
+optional<int> Parameters::branch_scale_modifiable_reg(int s) const
+{
+    return compute_expression_is_modifiable_reg(branch_scale_index(s));
+}
 
 void Parameters::branch_scale(int s, double x)
 {
-    auto R = compute_expression_is_modifiable_reg(branch_scale_index(s));
-
-    if (R)
+    if (auto R = branch_scale_modifiable_reg(s))
 	return set_modifiable_value(*R, x);
     else
 	throw myexception()<<"Branch scale "<<s+1<<" is not directly modifiable!";
@@ -1370,6 +1343,17 @@ vector<int> Parameters::partitions_for_imodel(int i) const
     return partitions;
 }
 
+vector<int> Parameters::partitions_for_scale(int i) const
+{
+    assert(0 <= i and i < n_branch_scales());
+    vector<int> partitions;
+    for(int p=0;p<n_data_partitions();p++)
+	if (auto index = scale_index_for_partition(p))
+	    if (*index == i)
+		partitions.push_back(p);
+    return partitions;
+}
+
 Parameters::Parameters(const std::shared_ptr<module_loader>& L,
 		       const vector<alignment>& A, const SequenceTree& tt,
 		       const vector<model_t>& SMs,
@@ -1394,6 +1378,7 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
     PC->constants.push_back(-1);
 
     add_modifiable_parameter_with_value("Heat.beta", 1.0);
+    variable_alignment_param = add_modifiable_parameter_with_value("*variable_alignment", variable_alignment_);
 
     // Add parameter for each scale
     vector<expression_ref> scales;
@@ -1402,7 +1387,7 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
 	string prefix = "Scale["+convertToString(i+1)+"]";
 
 	auto scale_model = scaleMs[i].expression;
-	scale_model = {var("Distributions.run_random'"), var("Prelude.Nothing"), 1.0, scale_model};
+	scale_model = {var("Distributions.gen_model_no_alphabet"), scale_model};
 	scale_model = {var("Distributions.do_log"), prefix, scale_model};
 	scale_model = {var("Prelude.unsafePerformIO"),scale_model};
 	scale_model = {var("Parameters.evaluate"),-1,scale_model};
@@ -1414,7 +1399,7 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
     add_parameter("Scale", scales_list);
 
     for(int i=0; i<n_branch_scales();i++)
-	PC->scale_parameter_indices[i] = add_compute_expression( {var("Prelude.!!"),scales_list,i} );
+	PC->scale_parameter_indices[i] = add_compute_expression( {var("Data.List.!!"),scales_list,i} );
 
     branches_from_affected_node.resize(tt.n_nodes());
 
@@ -1438,7 +1423,7 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
 	for(auto b: branch_list)
 	    branch_list_.push_back(b);
 
-	vector<int> b2 = evaluate_expression( {var("Prelude.listToVectorInt"),{var("Tree.edgesBeforeEdge"),my_tree(),b}}).as_<Vector<int>>();
+	vector<int> b2 = evaluate_expression( {var("Prelude.list_to_vector"),{var("Tree.edgesBeforeEdge"),my_tree(),b}}).as_<EVector>();
 	assert(b2.size() == branch_list_.size());
 	for( int i: branch_list_)
 	    assert(includes(b2,i));
@@ -1458,7 +1443,7 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
 	const alphabet& a = A[*first_index].get_alphabet();
 
 	expression_ref smodel = SMs[i].expression;
-	smodel = {var("Distributions.run_random'"), a, 1.0, smodel};
+	smodel = {var("Distributions.gen_model_with_alphabet"), a, smodel};
 	smodel = {var("Distributions.do_log"), prefix, smodel};
 	smodel = {var("Prelude.unsafePerformIO"),smodel};
 	smodel = {var("Parameters.evaluate"),-1,smodel};
@@ -1472,7 +1457,7 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
     {
 	string prefix = "I" + convertToString(i+1);
 	expression_ref imodel = IMs[i].expression;
-	imodel = {var("Distributions.run_random'"), var("Prelude.Nothing"), 1.0, imodel};
+	imodel = {var("Distributions.gen_model_no_alphabet"), imodel};
 	imodel = {var("Distributions.do_log"), prefix, imodel};
 	imodel = {var("Prelude.unsafePerformIO"),imodel};
 	imodel = {var("Parameters.evaluate"),-1,imodel};
@@ -1496,7 +1481,7 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
 	for(int b=0;b<t().n_branches();b++)
 	{
 	    expression_ref length = get_expression(TC->branch_duration_index[b]);
-	    int index = add_compute_expression( {var("Prelude.*"),scale,length} );
+	    int index = add_compute_expression( {var("Compiler.Num.*"),scale,length} );
 	    PC->branch_length_indices[s].push_back(index);
 	}
     }
@@ -1542,7 +1527,9 @@ Parameters::Parameters(const std::shared_ptr<module_loader>& L,
     }
 
     // create data partitions
-    bool allow_compression = load_value("compression",true);
+
+    // FIXME! Make likelihood_calculators for 1- and 2-sequence alignments handle compressed alignments.
+    bool allow_compression = load_value("site-compression", t().n_nodes() > 2);
 
     assert(like_calcs.size() == A.size());
     for(int i=0;i<A.size();i++)
@@ -1598,10 +1585,7 @@ bool accept_MH(const Model& P1,const Model& P2,log_double_t rho)
 	std::cerr<<endl<<endl;
     }
 
-    log_double_t p1 = P1.heated_probability();
-    log_double_t p2 = P2.heated_probability();
-
-    log_double_t ratio = rho*(p2/p1);
+    log_double_t ratio = rho*P2.heated_probability_ratio(P1);
 
     bool accept = (ratio >= 1.0 or uniform() < ratio);
 
