@@ -106,6 +106,11 @@ using std::valarray;
 using boost::program_options::variables_map;
 using boost::shared_ptr;
 
+expression_ref fst  = var("Data.Tuple.fst");
+expression_ref snd  = var("Data.Tuple.snd");
+expression_ref bind = var("Compiler.Base.>>=");
+expression_ref do_return = var("Compiler.Base.return");
+
 string model_t::show(bool top) const
 {
     if (top)
@@ -174,8 +179,10 @@ string default_markov_model(const alphabet& a)
 	return "lg08";
     else if (dynamic_cast<const Codons*>(&a))
 	return "gy94";
+    else if (dynamic_cast<const Doublets*>(&a))
+	return "tn93_sym+x2_sym+f";
     else if (dynamic_cast<const Triplets*>(&a))
-	return "x3[tn93_sym]";
+	return "tn93+x3";
     else
 	return "";
 }
@@ -340,7 +347,7 @@ expression_ref get_constant_model(const ptree& model_rep)
     if (expression_ref C = parse_constant(model_rep))
     {
 	if (model_rep.size() != 0) throw myexception()<<"An constant cannot have arguments!\n  '"<<model_rep.show()<<"'";
-	return {var("Prelude.return"), Tuple(C,List())};
+	return {do_return, Tuple(C,List())};
     }
     else
 	return {};
@@ -371,13 +378,21 @@ optional<pair<expression_ref,set<string>>> get_variable_model(const ptree& E, co
 	V = x;
 
     // 4. Construct the logging tuple and return it in order to allow this action to be performed.
-    V = {var("Prelude.return"),Tuple(V,List())};
+    V = {do_return, Tuple(V,List())};
 
     // 5. We need an extra level of {} to allow constructing the optional.
     return {{V,lambda_vars}};
 }
 
 
+/*
+ *
+ * do
+ *   pair_var <- var_body
+ *   let var_name = fst pair_var
+ *   pair_body <- let_body
+ *   return (fst pair_body, [("let:var",(Nothing,[(var_name,pair_x)])),("let:body",(Nothing,snd pair_body))])
+ */
 optional<pair<expression_ref,set<string>>> get_model_let(const Rules& R, const ptree& model_rep, const names_in_scope_t& scope)
 {
     auto name = model_rep.get_value<string>();
@@ -392,11 +407,28 @@ optional<pair<expression_ref,set<string>>> get_model_let(const Rules& R, const p
     var pair_x("pair_var_"+var_name);
     var x("var_"+var_name);
 
+    var pair_body("pair_body");
     var_info_t var_info(x,is_random(var_exp, scope));
 
     // 1. Perform the body with var_name in scope
     auto p = get_model_as(R, body_exp, extend_scope(scope, var_name, var_info));
-    expression_ref E = p.first;
+    expression_ref let_body = p.first;
+
+    // FIXME: we currently prohibit var_exp from containing any lambda-variables, so we don't need to check if it has them.
+    bool do_log = is_unlogged_random(R, var_exp, scope);
+    expression_ref var_loggers = List();
+    var_loggers = {var("Distributions.add_logger"), var_loggers, var_name, pair_x, do_log};
+
+    expression_ref loggers = List();
+    var Nothing("Data.Maybe.Nothing");
+    loggers = {var("Distributions.add_logger"),loggers,"let:body",Tuple(Nothing,{snd,pair_body}),false};
+    loggers = {var("Distributions.add_logger"),loggers,"let:var",Tuple(Nothing,var_loggers),false};
+
+    // E = return (fst pair_body, loggers)
+    expression_ref E = {do_return,Tuple({fst,pair_body},loggers)};
+
+    // E = do {pair_body <- let_body ; E}
+    E = {bind, let_body, lambda_quantify(pair_body,E)};
 
     // 2. Perform the variable expression
     {
@@ -405,10 +437,11 @@ optional<pair<expression_ref,set<string>>> get_model_let(const Rules& R, const p
 	if (arg_p.second.size())
 	    throw myexception()<<"You cannot let-bind a variable to an expression with a function-variable";
 
-	// E = 'arg <<= (\pair_var_name -> let {arg_name=fst pair_var_name} in E)
-	E = lambda_quantify(pair_x, let_expression({{x,{var("Prelude.fst"),pair_x}}},E));
+	// E = let arg_name=fst pair_var_name; E
+	E = let_expression({{x,{fst,pair_x}}},E);
 
-	E = {var("Prelude.>>="), arg, E};
+	// E = do {pair_x <- arg ; E}
+	E = {bind, arg, lambda_quantify(pair_x,E)};
     }
 
     return {{E, p.second}};
@@ -435,7 +468,7 @@ optional<pair<expression_ref,set<string>>> get_model_lambda(const Rules& R, cons
     auto lambda_vars = body_E.second;
 
     // E = E x l1 l2 l3
-    expression_ref E = {var("Prelude.fst"),pair_arg_body};
+    expression_ref E = {fst,pair_arg_body};
     for(auto& vname: lambda_vars)
 	E = {E, body_scope.at(vname).x};
 
@@ -450,10 +483,10 @@ optional<pair<expression_ref,set<string>>> get_model_lambda(const Rules& R, cons
 	E = lambda_quantify(scope.at(vname).x,E);
 
     // E = return $ (E,snd pair_arg_body)
-    E = {var("Prelude.return"),Tuple(E,{var("Prelude.snd"),var("pair_arg_body")})};
+    E = {do_return, Tuple(E,{snd,var("pair_arg_body")})};
 
     // E = do {pair_arg_body <- body ; E}
-    E = {var("Prelude.>>="),body,lambda_quantify(pair_arg_body,E)};
+    E = {bind,body,lambda_quantify(pair_arg_body,E)};
 
 
     // In summary, we have
@@ -505,7 +538,6 @@ pair<expression_ref,set<string>> get_model_function(const Rules& R, const ptree&
     bool perform_function = rule->get("perform",false);
     ptree call = rule->get_child("call");
     ptree args = rule->get_child("args");
-    var fst("Prelude.fst");
     
     if (not is_qualified_symbol(call.get_value<string>()) and not is_haskell_builtin_con_name(call.get_value<string>()))
 	throw myexception()<<"For rule '"<<name<<"', function '"<<call.get_value<string>()<<"' must be a qualified symbol or a builtin constructor like '(,)', but it is neither!";
@@ -564,11 +596,10 @@ pair<expression_ref,set<string>> get_model_function(const Rules& R, const ptree&
     }
 
     // 6. Return the function call: 'return (f call.name1 call.name2 call.name3)'
-    expression_ref Return = var("Prelude.return");
     if (not perform_function)
-	E = {Return,E};
+	E = {do_return,E};
 
-    E = {var("Prelude.>>="), E, lambda_quantify(var("result"),{Return,Tuple(var("result"),loggers)})};
+    E = {bind, E, lambda_quantify(var("result"),{do_return,Tuple(var("result"),loggers)})};
 
     // 7. Peform the rule arguments 'Prefix "arg_name" (arg_+arg_name) >>= (\arg_name -> (Log "arg_name" arg_name) << E)'
     for(int i=0;i<args.size();i++)
@@ -596,11 +627,11 @@ pair<expression_ref,set<string>> get_model_function(const Rules& R, const ptree&
 	if (arg_lambda_vars[i].empty())
 	{
 	    var x("arg_var_"+arg_name);
-	    E = let_expression({{x,{var("Prelude.fst"),pair_x}}},E);
+	    E = let_expression({{x,{fst,pair_x}}},E);
 	}
 	E = lambda_quantify(pair_x, E);
 
-	E = {var("Prelude.>>="), arg, E};
+	E = {bind, arg, E};
     }
 
     return {E, lambda_vars};
@@ -782,7 +813,7 @@ vector<pair<string, ptree>> extract_terms(ptree& m)
     return extracted;
 }
 
-#include "startup/help.hh"
+#include "util/text.H"
 
 string pretty_model_t::show_extracted() const
 {
