@@ -408,7 +408,7 @@ void help_topics(std::ostream& o, const ptree& help)
     auto subtopics = get_subtopics(help);
     subtopics.push_back("functions");
     
-    o<<"To help on one of the following topics, run `bali-phy help "<<underline("topic")<<"`\n\n";
+    o<<"To see help on one of the following topics, run `bali-phy help "<<underline("topic")<<"`\n\n";
     o<<show_options(subtopics);
     o<<"\n";
     o<<"\n";
@@ -440,15 +440,17 @@ void show_help(const string& topic, const vector<fs::path>& package_paths)
     Rules R(package_paths);
     if (topic == "functions")
     {
+	vector<string> func_names;
+
 	for(auto& rule: R.get_rules())
 	{
 	    if (rule.second.get_child_optional("name"))
 	    {
 		string name = rule.second.get_child("name").get_value<string>();
-		string result_type = unparse_type(rule.second.get_child("result_type"));
-		std::cout<<name<<" :: "<<result_type << std::endl;
+		func_names.push_back(name);
 	    }
 	}
+	std::cout<<show_options(func_names)<<std::endl;
 	return;
     }
 
