@@ -1,3 +1,4 @@
+#include <iostream>
 /*
   Copyright (C) 2004-2009 Benjamin Redelings
 
@@ -17,15 +18,18 @@
   along with BAli-Phy; see the file COPYING.  If not see
   <http://www.gnu.org/licenses/>.  */
 
+#include <algorithm>
+#include <map>
+#include <sstream>
+
+#include "myexception.H"
 #include "tree/tree.H"
 #include "util.H"
-#include <algorithm>
-#include <sstream>
-#include "myexception.H"
 
 using std::vector;
 using std::string;
 using std::pair;
+using std::map;
 using boost::any;
 using boost::dynamic_bitset;
 using boost::optional;
@@ -463,6 +467,92 @@ string get_attribute_string(const tree_attributes& attributes, const vector<stri
     return output;
 }
 
+int quoting_level_for_newick(const string& s)
+{
+    int level = 0;
+    for(auto c: s)
+    {
+	if (not isgraph(c))
+	{
+	    if (c != ' ') return 2;
+	    level = 1;
+	}
+	// Why no quotes needed for "("?
+	else if (strchr("(){}\"-]/\\,;:=*`+<>", c) != nullptr)
+	    return (s.size() > 1 ? 2 : 1);
+	else if (strchr("\'[_", c) != nullptr)
+	    return 2;
+    }
+    return level;
+}
+
+string quote_for_newick(const string& s)
+{
+    constexpr char squote = '\'';
+    string s2;
+    s2.reserve(s.size()+4);
+    s2.append(1,squote);
+    for(auto& c: s)
+    {
+	s2.append(1, c);
+	if (c == squote)
+	    s2.append(1, squote);
+    }
+    s2.append(1, squote);
+    return s2;
+}
+
+string unquote_for_newick(const string& s)
+{
+    assert(s[0] == '\'');
+    if (s.size() < 2 or s.back() != '\'')
+	throw myexception()<<"Quoted newick label \""<<s<<"\" doesn't end with single quote!";
+
+    string s2;
+    for(int i=1;i<s.size()-1;i++)
+    {
+	char c = s[i];
+	if (c == '\'' and i > 1 and s[i-1] == '\'') continue;
+	s2.append(1,c);
+    }
+    return s2;
+}
+
+string blanks_to_underscores(const string& s)
+{
+    string s2 = s;
+    std::replace(begin(s2), end(s2), ' ', '_');
+    return s2;
+}
+
+string underscores_to_blanks(const string& s)
+{
+    string s2 = s;
+    std::replace(begin(s2), end(s2), '_', ' ');
+    return s2;
+}
+
+string escape_for_newick(const string& s)
+{
+    int level = quoting_level_for_newick(s);
+    if (level == 0)
+	return s;
+    else if (level == 1)
+	return blanks_to_underscores(s);
+    else
+	return quote_for_newick(s);
+}
+
+string unescape_from_newick(const string& s)
+{
+    if (s.empty()) return s;
+
+    if (s[0] == '\'')
+	return unquote_for_newick(s);
+    else
+	return underscores_to_blanks(s);
+}
+
 string write(const vector<string>& names, 
 	     const vector<string>& node_attribute_names, const vector<string>& undirected_branch_attribute_names,
 	     const_branchview b, bool print_lengths)
@@ -490,7 +580,7 @@ string write(const vector<string>& names,
 
     // Print the name (it might be empty)
     if (names[n].size())
-	output += "'" + names[n] + "'";
+	output += escape_for_newick(names[n]);
 
     // Print the node attributes, if any
     output += get_attribute_string(n.attributes(), node_attribute_names, "&&NHX:", ":", node_label_index);
@@ -532,7 +622,7 @@ string write(const_nodeview root, const vector<string>& names,
 
     // Print the name (it might be empty)
     if (names[root].size())
-	output += "'" + names[root] + "'";
+	output += escape_for_newick(names[root]);
 
     // Print the node attributes, if any
     output += get_attribute_string(root.attributes(), node_attribute_names, "&&NHX:", ":", node_label_index);
@@ -728,6 +818,55 @@ void shift_leaves(BranchNode* start,int first,int n) {
     }
 }
 
+vector<int> Tree::prune_leaf(int n)
+{
+    assert(is_leaf_node(n));
+    BranchNode* leaf = nodes_[n];
+    int L = n_leaves();
+
+    // ensure that nodes_[i!=n] only point to surviving nodes
+    if (leaf->out)
+    {
+	int neighbor = leaf->out->node_attributes->name;
+	if (nodes_[neighbor] == leaf->out)
+	    nodes_[neighbor] = nodes_[neighbor]->next;
+	assert(::is_leaf_node(nodes_[neighbor]) or nodes_[neighbor] != leaf->out);
+    }
+
+    // get pointers to the current leaves that we will keep
+    vector<BranchNode*> old_nodes = nodes_;
+    old_nodes[n] = 0;
+
+    // remove the leaf
+    BranchNode* parent = TreeView::unlink_subtree(leaf->out);
+    TreeView(leaf).destroy();
+
+    // shifter down higher leaf names
+    for(int i=n+1;i<L;i++)
+	name_node(old_nodes[i],i-1);
+
+    // Reconstruct everything from leaf names
+    reanalyze(parent);
+
+    assert(n_leaves() == L-1);
+
+    // Construct the map from new to old node names
+    vector<int> mapping(n_nodes(), -1);
+    for(int i=0;i<old_nodes.size();i++)
+	if (old_nodes[i])
+	    mapping[old_nodes[i]->node_attributes->name] = i;
+
+    // Check that the mapping for each new name is defined, and unique
+    for(int i=0;i<mapping.size();i++)
+    {
+	assert(mapping[i] != -1);
+	for(int j=0;j<i;j++)
+	    assert(mapping[i] != mapping[j]);
+    }
+
+    return mapping;
+}
+
 /// Remove the specified leaves and their dangling ancestors
 vector<int> Tree::prune_leaves(const vector<int>& remove) 
 {
@@ -1836,27 +1975,29 @@ void add_comments(vector< pair<string,any> >& tags, const vector<string>& commen
 
 #include <iostream>
 
-bool has_label(BranchNode* BN, int node_label_index)
+bool has_label(BranchNode* BN, optional<int> node_label_index)
 {
-    if (node_label_index == -1) return false;
+    if (not node_label_index) return false;
 
-    const auto& label_any = (*BN->node_attributes)[node_label_index];
+    const auto& node_label = (*BN->node_attributes)[*node_label_index];
 
-    if (boost::any_cast<string>(&label_any))
-	return true;
-    else
-	return false;
+    return not node_label.empty();
 }
 
-const string& get_label(BranchNode* BN, int node_label_index)
+string get_label(BranchNode* BN, optional<int> node_label_index)
 {
-    assert(node_label_index != -1);
+    if (not has_label(BN, node_label_index)) return "";
 
-    const auto& label_any = (*BN->node_attributes)[node_label_index];
+    const auto& label_any = (*BN->node_attributes)[*node_label_index];
 
     return boost::any_cast<const string&>(label_any);
 }
 
+bool has_non_empty_label(BranchNode* BN, optional<int> node_label_index)
+{
+    return get_label(BN, node_label_index).size() > 0;
+}
+
 /*
  * Tree -> Branch ;
  * Branch -> [Node] [string] [: double]
@@ -1868,14 +2009,14 @@ const string& get_label(BranchNode* BN, int node_label_index)
 int Tree::parse_(const string& line, std::function<void(BranchNode*)> assign_names)
 {
     node_attribute_names.clear();
-    if (node_label_index != -1)
-	node_attribute_names.resize(node_label_index+1);
+    if (node_label_index)
+	node_attribute_names.resize(*node_label_index+1);
 
     directed_branch_attribute_names.clear();
 
     undirected_branch_attribute_names.clear();
-    if (branch_length_index != -1)
-	undirected_branch_attribute_names.resize(branch_length_index+1);
+    if (branch_length_index)
+	undirected_branch_attribute_names.resize(*branch_length_index+1);
 
     const string delimiters = "(),:;";
     const string whitespace = "\t\n ";
@@ -1926,6 +2067,8 @@ int Tree::parse_(const string& line, std::function<void(BranchNode*)> assign_nam
 	}
 	else if (word == ",")
 	{
+	    if (tree_stack.size() <= 1)
+		throw myexception()<<"Reading tree: found ',' outside parenthesis!";
 	    if (pos == 0 or pos == 1 or pos == 2)
 		set_attributes(tags, node_attribute_names, *BN->node_attributes);
 	    else if (pos == 3 or pos == 4)
@@ -1952,10 +2095,19 @@ int Tree::parse_(const string& line, std::function<void(BranchNode*)> assign_nam
 	}
 	else if (word == ":")
 	{
-	    set_attributes(tags, node_attribute_names, *BN->node_attributes);
-
 	    if (pos > 2)
 		throw myexception()<<"Cannot have a ':' here! (pos == "<<pos<<")";
+
+	    // Handle (a,b):1.0; -- we should ignore any attributes for the branch in this case
+	    if (tree_stack.size() > 1)
+		set_attributes(tags, node_attribute_names, *BN->node_attributes);
+	    else
+	    {
+		// There should be only a single node, or we'd be in a situation like a,b:1.0;
+		// If we have a:1.0,b:1.0; the we will complain about the ",", so treating the a this way should be safe;
+		assert(tree_stack.back().size() == 1);
+	    }
+
 	    pos = 3;
 	}
 	else
@@ -1964,8 +2116,8 @@ int Tree::parse_(const string& line, std::function<void(BranchNode*)> assign_nam
 	    {
 		set_attributes(tags, node_attribute_names, *BN->node_attributes);
 
-		if (node_label_index != -1)
-		    (*BN->node_attributes)[node_label_index] = word;
+		if (node_label_index)
+		    (*BN->node_attributes)[*node_label_index] = unescape_from_newick(word);
 
 		pos = 2;
 	    }
@@ -1973,10 +2125,14 @@ int Tree::parse_(const string& line, std::function<void(BranchNode*)> assign_nam
 		throw myexception()<<"Node name '"<<word<<"' comes directly after '"<<prev<<"'";
 	    else if (pos == 3)
 	    {
-		set_attributes(tags, undirected_branch_attribute_names, *BN->undirected_branch_attributes);
+		// Handle (a,b):1.0; -- we should ignore any attributes for the branch in this case
+		if (tree_stack.size() > 1)
+		{
+		    set_attributes(tags, undirected_branch_attribute_names, *BN->undirected_branch_attributes);
 
-		if (branch_length_index != -1)
-		    (*BN->undirected_branch_attributes)[branch_length_index] = convertTo<double>(word);	
+		    if (branch_length_index)
+			(*BN->undirected_branch_attributes)[*branch_length_index] = convertTo<double>(word);
+		}
 		pos = 4;
 	    }
 	    else if (pos == 4)
@@ -2012,7 +2168,7 @@ int Tree::parse_and_discover_names(const string& line)
 	// First give integer name to leaves with labels.
 	for(BN_iterator BN(root_);BN;BN++)
 	{
-	    if (::is_leaf_node(*BN) and has_label(*BN, node_label_index))
+	    if (::is_leaf_node(*BN) and has_non_empty_label(*BN, node_label_index))
 		(*BN)->node_attributes->name = L++;
 	    else
 		(*BN)->node_attributes->name = -1;
@@ -2021,7 +2177,7 @@ int Tree::parse_and_discover_names(const string& line)
 	// Name other leaves and resize attribute objects
 	for(BN_iterator BN(root_);BN;BN++)
 	{
-	    if (::is_leaf_node(*BN) and not has_label(*BN, node_label_index))
+	    if (::is_leaf_node(*BN) and not has_non_empty_label(*BN, node_label_index))
 		(*BN)->node_attributes->name = L++;
 
 	    (*BN)->node_attributes->resize(n_node_attributes());
@@ -2036,41 +2192,59 @@ int Tree::parse_and_discover_names(const string& line)
     return parse_(line, namer);
 }
 
-int get_leaf_index(const string& word, bool allow_numbers, const vector<string>& names)
+int get_leaf_index(const string& word, bool allow_numbers, const map<string,optional<int>>& name_to_index)
 {
     auto leaf_index = can_be_converted_to<int>(word);
     if (allow_numbers and leaf_index)
     {
 	if (*leaf_index < 1)
 	    throw myexception()<<"Leaf index '"<<word<<"' is negative: not allowed!";
-	if (*leaf_index > names.size())
-	    throw myexception()<<"Leaf index '"<<word<<"' is too high: the taxon set contains only "<<names.size()<<" taxa.";
+	if (*leaf_index > name_to_index.size())
+	    throw myexception()<<"Leaf index '"<<word<<"' is too high: the taxon set contains only "<<name_to_index.size()<<" taxa.";
 	return *leaf_index - 1;
     }
-    else if (names.size() == 0)
-	throw myexception()<<"Leaf name '"<<word<<"' is not an integer!";
+    else if (name_to_index.size() == 0)
+    {
+	if (allow_numbers)
+	    throw myexception()<<"Leaf name '"<<word<<"' is not an integer!";
+	else
+	    throw myexception()<<"Leaf name '"<<word<<"' is an integer, but integers are not allowed!";
+    }
     else 
     {
-	leaf_index = find_index(names,word);
-	if (leaf_index)
+	if (not name_to_index.count(word))
+	    throw myexception()<<"Leaf name '"<<word<<"' is not in the specified taxon set!";
+	else if (auto leaf_index = name_to_index.at(word))
 	    return *leaf_index;
 	else
-	    throw myexception()<<"Leaf name '"<<word<<"' is not in the specified taxon set!";
+	    throw myexception()<<"Leaf name '"<<word<<"' does not uniquely specify a leaf - multiple leaves have this name!";
     }
 }
 
+// This approach assumes that all leaf names are unique??
 int Tree::parse_with_names_or_numbers(const string& line,const vector<string>& names,bool allow_numbers)
 {
     if (names.size() == 0 and not allow_numbers)
 	throw myexception()<<"Tree::parse_with_names_or_numbers( ): must supply leaf names if integers are not allowed.";
 
-    auto namer = [this, allow_numbers, &names](BranchNode* root_)
+    // Make sure that we're not trying to identify nodes by non-unique names
+    map<string,optional<int>> name_to_index;
+    for(int i=0; i<names.size();i++)
+    {
+	auto& name = names[i];
+	if (name_to_index.count(name))
+	    name_to_index[name] = boost::none;
+	else
+	    name_to_index[name] = i;
+    }
+
+    auto namer = [this, allow_numbers, &name_to_index](BranchNode* root_)
     {
 	// Name leaves and resize attribute vectors
 	for(BN_iterator BN(root_);BN;BN++)
 	{
-	    if (::is_leaf_node(*BN) and has_label(*BN, node_label_index))
-		(*BN)->node_attributes->name = get_leaf_index(get_label(*BN, node_label_index), allow_numbers,names);
+	    if (::is_leaf_node(*BN))
+		(*BN)->node_attributes->name = get_leaf_index(get_label(*BN, node_label_index), allow_numbers, name_to_index);
 	    else
 		(*BN)->node_attributes->name = -1;
 
@@ -2130,16 +2304,25 @@ vector<int> Tree::branches_before(int b) const
     
 }
 
+vector<int> Tree::branches_after(int b) const
+{
+    vector<int> bb;
+    for(auto bv = directed_branch(b).branches_after();bv;bv++)
+	bb.push_back((*bv).name());
+    return bb;
+    
+}
+
 Tree::Tree()
     :caches_valid(false),
      n_leaves_(0),
      node_label_index(0),
      branch_length_index(0)
 {
-    if (node_label_index != -1 and node_label_index >= n_node_attributes())
-	set_n_node_attributes(node_label_index+1);
-    if (branch_length_index != -1 and branch_length_index >= n_undirected_branch_attributes())
-	set_n_undirected_branch_attributes(branch_length_index+1);
+    if (node_label_index and *node_label_index >= n_node_attributes())
+	set_n_node_attributes(*node_label_index+1);
+    if (branch_length_index and *branch_length_index >= n_undirected_branch_attributes())
+	set_n_undirected_branch_attributes(*branch_length_index+1);
 }
 
 Tree::Tree(BranchNode* BN) 
@@ -2153,10 +2336,10 @@ Tree::Tree(BranchNode* BN)
     if (BN->directed_branch_attributes)
 	directed_branch_attribute_names.resize( BN->directed_branch_attributes->size() );
 
-    if (node_label_index != -1 and node_label_index >= n_node_attributes())
-	set_n_node_attributes(node_label_index+1);
-    if (branch_length_index != -1 and branch_length_index >= n_undirected_branch_attributes())
-	set_n_undirected_branch_attributes(branch_length_index+1);
+    if (node_label_index  and *node_label_index >= n_node_attributes())
+	set_n_node_attributes(*node_label_index+1);
+    if (branch_length_index and *branch_length_index >= n_undirected_branch_attributes())
+	set_n_undirected_branch_attributes(*branch_length_index+1);
 
     reanalyze(BN);
 }
@@ -2242,6 +2425,19 @@ void RootedTree::remove_node_from_branch(int node)
     Tree::remove_node_from_branch(node);
 }
 
+vector<int> RootedTree::prune_leaf(int n) 
+{
+    if (root_ and root_->node_attributes->name == n)
+	root_ = NULL;
+    if (root_ and root_->out and root_->out->node_attributes->name == n)
+	root_ = root_->next;
+    if (root_)
+	assert(::is_leaf_node(root_) or not root_->out or root_->out->node_attributes->name != n);
+    
+    return Tree::prune_leaf(n);
+}
+
+
 vector<int> RootedTree::prune_leaves(const vector<int>& remove) 
 {
     root_ = NULL;
@@ -2417,13 +2613,56 @@ boost::dynamic_bitset<> branch_partition(const Tree& T,int b)
 }
 
 
-double tree_length(const Tree& T) {
+double tree_length(const Tree& T)
+{
     double total = 0;
     for(int i=0;i<T.n_branches();i++)
 	total += T.branch(i).length();
     return total;
 }
 
+double tree_diameter(const Tree& T)
+{
+    vector<optional<double>> length_behind_inclusive(T.n_branches()*2);
+
+    // Start of handling leaf branches, with nothing behind them.
+    vector<int> branches_to_process;
+    for(int i=0; i<T.n_leafbranches(); i++)
+	branches_to_process.push_back(T.directed_branch(T.leaf_branch(i)));
+
+    for(int i=0; i<branches_to_process.size();i++)
+    {
+	int b = branches_to_process[i];
+
+	double d = 0;
+	for(int b2: T.branches_before(b))
+	    d = std::max(d, *length_behind_inclusive[b2]);
+	length_behind_inclusive[b] = d + T.directed_branch(b).length();
+
+	for(int b2: T.branches_after(b))
+	{
+	    bool ready = true;
+	    for(int b3: T.branches_before(b2))
+		ready = ready and length_behind_inclusive[b3];
+
+	    if (ready)
+		branches_to_process.push_back(b2);
+	}
+    }
+
+    for(int i=0; i<2*T.n_branches();i++)
+	assert(length_behind_inclusive[i]);
+
+    double D = 0;
+    for(int i=0; i<T.n_leafbranches(); i++)
+    {
+	int b = T.directed_branch(T.leaf_branch(i).reverse());
+	D = std::max(D, *length_behind_inclusive[b]);
+    }
+
+    return D;
+}
+
 int subtree_height(const Tree& T,int b) 
 {
     int node = T.directed_branch(b).target();
